import argparse
import re
from typing import Callable, Iterable, List, NoReturn, Optional, Set, Union

import pandas as pd

from mmproteo.utils import log
from mmproteo.utils.config import Config


class AbstractFilterConditionNode:
    def __call__(self, row: pd.Series) -> Optional[bool]:
        raise NotImplementedError


class AndFilterConditionNode(AbstractFilterConditionNode):
    def __init__(self, conditions: List[AbstractFilterConditionNode]):
        super(AndFilterConditionNode, self).__init__()
        self.conditions = conditions

    def __call__(self, row: pd.Series) -> Optional[bool]:
        res = None
        for condition in self.conditions:
            inner_res = condition(row=row)
            if inner_res is not None:
                if not inner_res:
                    return False
                else:
                    res = True
        return res


class OrFilterConditionNode(AbstractFilterConditionNode):
    def __init__(self, conditions: List[AbstractFilterConditionNode]):
        super(OrFilterConditionNode, self).__init__()
        self.conditions = conditions

    def __call__(self, row: pd.Series) -> Optional[bool]:
        res = None
        for condition in self.conditions:
            inner_res = condition(row=row)
            if inner_res is not None:
                if inner_res:
                    return True
                else:
                    res = False
        return res


class NotFilterConditionNode(AbstractFilterConditionNode):
    def __init__(self, condition: AbstractFilterConditionNode):
        super(NotFilterConditionNode, self).__init__()
        self.condition = condition

    def __call__(self, row: pd.Series) -> Optional[bool]:
        res = self.condition(row=row)
        if res is None:
            return None
        return not res


class ColumnRegexFilterConditionNode(AbstractFilterConditionNode):
    def __init__(self, column_name: str, value_regex: str):
        super(ColumnRegexFilterConditionNode, self).__init__()
        self.column_name = column_name
        self.value_regex = re.compile(value_regex)

    def __call__(self, row: pd.Series) -> Optional[bool]:
        if self.column_name in row.index:
            return bool(self.value_regex.fullmatch(str(row[self.column_name])))
        return None


class NoneFilterConditionNode(AbstractFilterConditionNode):
    def __init__(self, condition: AbstractFilterConditionNode, none_value: bool = True):
        super(NoneFilterConditionNode, self).__init__()
        self.condition = condition
        self.none_value = none_value

    def __call__(self, row: pd.Series) -> bool:
        res = self.condition(row=row)
        if res is None:
            res = self.none_value
        return res


def create_or_filter_from_str(or_filter_str: str) -> Union[OrFilterConditionNode, NoReturn]:
    conditions = []

    column_condition_strings = or_filter_str.split(Config.default_filter_or_separator)
    for column_condition_string in column_condition_strings:
        parts = re.split(pattern=Config.default_filter_separator_regex,
                         string=column_condition_string,
                         maxsplit=1)
        if len(parts) != 2 or len(parts[0]) == 0 or len(parts[1]) == 0:
            raise argparse.ArgumentTypeError(f"'{column_condition_string}' does not match the column filter pattern "
                                             f"'.+{Config.default_filter_separator_regex}.+'")
        column_name = parts[0]
        value_regex = parts[1]
        negation = column_condition_string[len(parts[0])] == "!"

        column_condition = ColumnRegexFilterConditionNode(column_name=column_name, value_regex=value_regex)

        if negation:
            column_condition = NotFilterConditionNode(condition=column_condition)

        conditions.append(column_condition)

    or_condition = OrFilterConditionNode(conditions=conditions)
    return or_condition


def create_file_extension_filter(required_file_extensions: Iterable[str],
                                 optional_file_extensions: Optional[Iterable[str]] = None) \
        -> Callable[[Optional[str]], bool]:
    if optional_file_extensions is None:
        file_extensions = set(required_file_extensions)
    else:
        file_extensions = {required_extension + "." + optional_extension
                           for required_extension in required_file_extensions
                           for optional_extension in optional_file_extensions}
        file_extensions.update(required_file_extensions)  # add all

    def filter_file_extension(filename: Optional[str]) -> bool:
        if filename is None:
            return True
        return filename.lower().endswith(tuple(file_extensions))

    return filter_file_extension


def filter_files_df(files_df: Optional[pd.DataFrame],
                    file_name_column: str = Config.default_file_name_column,
                    file_extensions: Optional[Union[List[str], Set[str]]] = None,
                    column_filter: Optional[AbstractFilterConditionNode] = None,
                    max_num_files: Optional[int] = None,
                    sort: bool = Config.default_filter_sort,
                    logger: log.Logger = log.DEFAULT_LOGGER) -> Optional[pd.DataFrame]:
    if files_df is None:
        return None

    if file_extensions is None or len(file_extensions) == 0:
        logger.debug("Skipping file extension filtering")
    else:
        logger.assert_true(file_name_column in files_df.columns, "Could not find '%s' column in files_df columns" %
                           file_name_column)
        required_file_extensions = file_extensions

        from mmproteo.utils.formats.archives import get_extractable_file_extensions
        optional_file_extensions = get_extractable_file_extensions()
        optional_file_extensions -= set(required_file_extensions)

        required_file_extensions_list_str = "\", \"".join(sorted(required_file_extensions))
        optional_file_extensions_list_str = "\", \"".join(sorted(optional_file_extensions))
        if len(optional_file_extensions_list_str) > 0:
            optional_file_extensions_list_str = " and the following optional file extensions " \
                                                f"[\"{optional_file_extensions_list_str}\"]"
        logger.info("Filtering files based on the following required file extensions "
                    f"[\"{required_file_extensions_list_str}\"]" + optional_file_extensions_list_str)

        file_extension_filter = create_file_extension_filter(required_file_extensions, optional_file_extensions)
        files_df = files_df[files_df[file_name_column].apply(file_extension_filter)]

        logger.debug("File extension filtering resulted in %d valid file names" % len(files_df))

    if column_filter is not None:
        column_filter = NoneFilterConditionNode(condition=column_filter, none_value=True)
        files_df = files_df[files_df.apply(func=column_filter, axis=1)]

    if sort:
        # sort, such that files with same prefixes but different extensions come in pairs
        files_df = files_df.sort_values(by=file_name_column)

    if max_num_files is not None and max_num_files > 0:
        files_df = files_df[:max_num_files]

    return files_df


def filter_files_list(filenames: List[Optional[str]],
                      file_extensions: Optional[Iterable[str]] = None,
                      column_filter: Optional[AbstractFilterConditionNode] = None,
                      max_num_files: Optional[int] = None,
                      keep_null_values: bool = Config.default_keep_null_values,
                      sort: bool = Config.default_filter_sort,
                      drop_duplicates: bool = Config.default_filter_drop_duplicates,
                      logger: log.Logger = log.DEFAULT_LOGGER) -> List[Optional[str]]:
    """

    :param column_filter:
    :param filenames:
    :param file_extensions:
    :param max_num_files:
    :param keep_null_values:    whether to keep null values in the given :param:`filenames`. Cannot be used
                                with :param:`drop_duplicates`
    :param sort:
    :param drop_duplicates:     whether to drop duplicate entries in :param:`filenames`. Cannot be used with
                                :param:`keep_null_values`.
    :param logger:
    :return:
    """
    logger.assert_true(not keep_null_values or not drop_duplicates,
                       "Cannot use keep_null_values and drop_duplicates simultaneously")

    if not keep_null_values:
        filenames = [filename for filename in filenames if filename is not None]
    if len(filenames) == 0:
        return list()

    df = pd.DataFrame(data=filenames, columns=[Config.default_file_name_column])
    if drop_duplicates:
        # TODO this also drops duplicate None values
        # it would require further preprocessing to use it together with keep_null_values
        # possible solution: copy the fileName column to a temporary column, fill all None values there with unique
        # values, drop_duplicates on this temporary column, remove the temporary column
        df = df.drop_duplicates()

    filtered_df = filter_files_df(files_df=df,
                                  file_name_column=Config.default_file_name_column,
                                  file_extensions=file_extensions,
                                  column_filter=column_filter,
                                  max_num_files=max_num_files,
                                  sort=sort,
                                  logger=logger)
    return filtered_df[Config.default_file_name_column].to_list()


import json
import os
from typing import Dict, Iterable, List, Optional, Union, NoReturn

import pandas as pd

import mmproteo.utils.filters
from mmproteo.utils import download as dl, log, utils
from mmproteo.utils.config import Config
from mmproteo.utils.download import AbstractDownloader
from mmproteo.utils.visualization import pretty_print_json


class AbstractPrideApi(AbstractDownloader):
    LIST_PROJECT_FILES_URL = None
    GET_PROJECT_SUMMARY_URL = None

    def __init__(self, version: str, logger: log.Logger = log.DEFAULT_LOGGER):
        super().__init__(logger)
        self.version = version

    def get_project_summary(self, project_name: str) -> Optional[dict]:
        project_summary_link = self.GET_PROJECT_SUMMARY_URL % project_name
        return self.request_json(project_summary_link, "project summary using API version " + self.version,
                                 self.logger)

    def get_project_files(self, project_name: str):
        raise NotImplementedError


class PrideApiV1(AbstractPrideApi):
    LIST_PROJECT_FILES_URL = "https://www.ebi.ac.uk/pride/ws/archive/file/list/project/%s"
    GET_PROJECT_SUMMARY_URL = "https://www.ebi.ac.uk/pride/ws/archive/project/%s"

    def get_project_files(self, project_name: str) -> Optional[pd.DataFrame]:
        project_files_link = self.LIST_PROJECT_FILES_URL % project_name
        response_dict = self.request_json(url=project_files_link,
                                          subject_name="list of project files using API version " + self.version,
                                          logger=self.logger)
        if response_dict is None:
            return None
        try:
            files_df = pd.DataFrame(pd.json_normalize(response_dict['list']))
            return files_df
        except Exception:
            return None


class PrideApiV2(AbstractPrideApi):
    LIST_PROJECT_FILES_URL = "https://www.ebi.ac.uk/pride/ws/archive/v2/files/byProject?accession=%s"
    GET_PROJECT_SUMMARY_URL = "https://www.ebi.ac.uk/pride/ws/archive/v2/projects/%s"

    @staticmethod
    def _get_ftp_or_http_file_location(file_location_entries: List[Dict[str, str]]) -> Optional[Dict[str, str]]:
        compatible_file_location_entries = [entry for entry in file_location_entries
                                            if 'value' in entry
                                            and entry['value'].lower().startswith(('ftp://', 'http://', 'https://'))]
        if len(compatible_file_location_entries) > 0:
            return compatible_file_location_entries[0]
        return None

    @staticmethod
    def _format_file_entry(file_entry: dict) -> dict:
        if 'publicFileLocations' in file_entry:
            try:
                file_entry['publicFileLocations'] = PrideApiV2._get_ftp_or_http_file_location(
                    file_entry['publicFileLocations'])
                file_entry['publicFileLocation'] = file_entry.pop('publicFileLocations')  # rename to singular
            except Exception:
                pass
        return file_entry

    def get_project_files(self,
                          project_name: str,
                          download_link_column: str = Config.default_download_link_column) -> Optional[pd.DataFrame]:
        project_files_link = self.LIST_PROJECT_FILES_URL % project_name
        response_dict_list = self.request_json(url=project_files_link,
                                               subject_name="list of project files using API version " + self.version,
                                               logger=self.logger)
        if response_dict_list is None:
            return None
        response_dict_list = [self._format_file_entry(response_dict) for response_dict in response_dict_list]

        try:
            files_df = pd.DataFrame(pd.json_normalize(response_dict_list))
        except Exception:
            return None

        files_df = files_df.rename(columns={"publicFileLocation.value": download_link_column})
        return files_df


PRIDE_APIS = {
    "1": PrideApiV1,
    "2": PrideApiV2,
}

DEFAULT_PRIDE_API_VERSIONS = ["2", "1"]


def get_pride_api_versions() -> List[str]:
    return sorted(PRIDE_APIS.keys())


def get_string_of_pride_api_versions(extension_quote: str = '"', separator: str = ", ") -> str:
    return utils.concat_set_of_options(options=DEFAULT_PRIDE_API_VERSIONS,
                                       option_quote=extension_quote,
                                       separator=separator)


def _query_project_summary(project_name: str,
                           api_versions: List[str] = None,
                           logger: log.Logger = log.DEFAULT_LOGGER) -> Optional[dict]:
    """Get the project as a json and return it as a dataframe"""
    if api_versions is None or len(api_versions) == 0:
        api_versions = DEFAULT_PRIDE_API_VERSIONS

    for api_version in api_versions:
        api = PRIDE_APIS[api_version](api_version, logger)
        response_dict = api.get_project_summary(project_name)
        if response_dict is not None:
            logger.info("Received project summary using API version %s for project \"%s\"" %
                        (api_version, project_name))
            return response_dict

    logger.warning("Could not get project summary.")
    return None


def get_project_info(project_name: str,
                     api_versions: List[str] = None,
                     logger: log.Logger = log.DEFAULT_LOGGER) -> Optional[str]:
    summary_dict = _query_project_summary(project_name=project_name, api_versions=api_versions, logger=logger)
    if summary_dict is None:
        return None
    return pretty_print_json(summary_dict)


def get_project_files(project_name: str,
                      api_versions: List[str] = None,
                      logger: log.Logger = log.DEFAULT_LOGGER) -> Optional[pd.DataFrame]:
    """Get the project as a json and return it as a dataframe"""
    if api_versions is None or len(api_versions) == 0:
        api_versions = DEFAULT_PRIDE_API_VERSIONS

    for api_version in api_versions:
        api = PRIDE_APIS[api_version](api_version, logger)
        files_df = api.get_project_files(project_name)
        if files_df is not None:
            logger.info("Received project file list using API version %s with %d files for project \"%s\"" %
                        (api_version, len(files_df), project_name))
            return files_df

    logger.warning("Could not get list of project files.")
    return None


def download(project_name: Optional[str] = None,
             project_files: Optional[pd.DataFrame] = None,
             valid_file_extensions: Optional[Iterable[str]] = None,
             max_num_files: Optional[int] = None,
             column_filter: Optional[mmproteo.utils.filters.AbstractFilterConditionNode] = None,
             download_dir: str = Config.default_storage_dir,
             skip_existing: bool = Config.default_skip_existing,
             count_failed_files: bool = Config.default_count_failed_files,
             count_skipped_files: bool = Config.default_count_skipped_files,
             file_name_column: str = Config.default_file_name_column,
             download_link_column: str = Config.default_download_link_column,
             downloaded_files_column: str = Config.default_downloaded_files_column,
             api_versions: List[str] = None,
             thread_count: int = Config.default_thread_count,
             logger: log.Logger = log.DEFAULT_LOGGER) \
        -> Union[Optional[pd.DataFrame], NoReturn]:
    logger.assert_true(condition=(project_name is not None or project_files is not None),
                       error_msg="either project_name or project_files are required")

    if project_files is None:
        project_files = get_project_files(project_name=project_name,
                                          api_versions=api_versions,
                                          logger=logger)
        if project_files is None:
            return None

    filtered_files = mmproteo.utils.filters.filter_files_df(files_df=project_files,
                                                            file_name_column=file_name_column,
                                                            file_extensions=valid_file_extensions,
                                                            max_num_files=None,
                                                            column_filter=column_filter,
                                                            sort=True,
                                                            logger=logger)
    logger.assert_true(download_link_column in filtered_files.columns,
                       f"Could not find column '{download_link_column}' in filtered_files dataframe")

    initial_directory = os.getcwd()
    os.chdir(download_dir)
    downloaded_files = dl.download_files(download_urls=filtered_files[download_link_column],
                                         skip_existing=skip_existing,
                                         count_failed_files=count_failed_files,
                                         count_skipped_files=count_skipped_files,
                                         max_num_files=max_num_files,
                                         keep_null_values=True,
                                         thread_count=thread_count,
                                         logger=logger)

    filtered_files = filtered_files.head(len(downloaded_files))
    filtered_files[downloaded_files_column] = downloaded_files
    os.chdir(initial_directory)

    return filtered_files



import os

from .utils.fixtures import run_in_temp_directory
from .utils.defaults import DEFAULT_PROJECT

MZ_PIPELINE_COMMAND = f"mmproteo -p {DEFAULT_PROJECT} -e mzid,mzml -n 2 -c fileName,fileSizeBytes " \
                      f"list download extract mz2parquet"


def test_mmproteo_download(run_in_temp_directory):
    os.system(MZ_PIPELINE_COMMAND)
    # TODO



import os
import subprocess
import time

try:
    from subprocess import DEVNULL  # Python 3.
except ImportError:
    DEVNULL = open(os.devnull, 'wb')
from typing import Any, Hashable, Iterable, List, Optional, Union, Dict, NoReturn, Callable

import numpy as np
import pandas as pd

from mmproteo.utils import log


def concat_set_of_options(options: Iterable[str], option_quote: str = '"', separator: str = ", ") -> str:
    if type(options) != list:
        options = sorted(options)
    return separator.join([option_quote + option + option_quote for option in options])


def deduplicate_list(lst: List[Hashable]) -> List[Hashable]:
    already_inserted = set()
    deduplicated_list = []
    for e in lst:
        if e not in already_inserted:
            already_inserted.add(e)
        deduplicated_list.append(e)
    return deduplicated_list


def _denumpyfy(element: Any) -> Any:
    if type(element) == np.int64:
        return int(element)
    if type(element) == np.float64:
        return float(element)
    if type(element) in [int, str, float]:
        return element
    if type(element) == dict:
        return {k: _denumpyfy(v) for k, v in element.items()}
    if type(element) == list:
        return [_denumpyfy(v) for v in element]
    if type(element) == set:
        return {_denumpyfy(v) for v in element}
    raise NotImplementedError(type(element))


def denumpyfy(element: Union[np.int64, np.float64, int, str, float, dict, list, set]) \
        -> Union[int, str, float, dict, list, set]:
    return _denumpyfy(element)


def ensure_dir_exists(directory: str, logger: log.Logger = log.DEFAULT_LOGGER) -> None:
    """Ensure the existence of a given directory (path) by creating it/them if they do not exist yet."""
    if len(directory) == 0:
        return
    try:
        os.makedirs(directory, exist_ok=True)
    except FileExistsError:
        logger.warning("'%s' already exists and is not a directory")


def flatten_single_element_containers(elem: Union[Iterable, Any]) -> Union[Iterable, Any]:
    try:
        # skip one-element lists or sets,...
        while type(elem) == list or type(elem) == set or type(elem) == tuple:
            if len(elem) == 1:
                elem = next(iter(elem))
                continue
            non_null_elements = [e for e in elem if e is not None]
            if len(non_null_elements) == 0:
                elem = None
                break
            if len(non_null_elements) == 1:
                elem = non_null_elements[0]
                continue
            break
    except Exception:
        pass
    return elem


def flatten_dict(input_dict: dict,
                 result_dict: dict = None,
                 overwrite: bool = False,
                 clean_keys: bool = True,
                 concat_keys: bool = True) -> dict:
    """

    :param input_dict:
    :param result_dict:
    :param overwrite:
    :param clean_keys:
    :param concat_keys:  Side effect is that all keys become strings.
    :return:
    """
    if result_dict is None:
        result_dict = dict()

    dict_queue = [("", input_dict)]

    while len(dict_queue) > 0:
        key_prefix, item = dict_queue[0]
        dict_queue = dict_queue[1:]

        for key, value in item.items():
            if clean_keys:
                key = str(key)
                key = key.replace(" ", "_")
                key = "".join([c for c in key if c.isalnum() or c == "_"])
            value = flatten_single_element_containers(value)

            if type(value) == dict:
                new_prefix = str(key) + "__"
                if concat_keys:
                    new_prefix = key_prefix + new_prefix
                dict_queue.append((new_prefix, value))
                continue

            if not overwrite and key not in result_dict:
                if concat_keys:
                    new_key = key_prefix + str(key)
                else:
                    new_key = key
                result_dict[new_key] = value

    return result_dict


def get_plural_s(count: int) -> str:
    if count > 1:
        return "s"
    else:
        return ""


def list_of_dicts_to_dict(items: List[Dict], dict_key: str) -> Union[Optional[Dict], NoReturn]:
    all_have_key = True

    for item in items:
        if dict_key not in item.keys():
            all_have_key = False
            break

    if all_have_key:
        items_dict = dict()
        for item in items:
            key = item[dict_key]
            items_dict[key] = item
        return items_dict

    return None


def format_command_template(command_template: str, formatter: Callable[[str], str]) -> List[str]:
    parts = command_template.split(" ")
    parts = [formatter(part) for part in parts]
    return parts


def stop_docker_container(container_name: str,
                          docker_stop_container_command_template: str = "docker stop {container_name}",
                          logger: log.Logger = log.DEFAULT_LOGGER) -> str:
    stop_command = format_command_template(docker_stop_container_command_template,
                                           lambda s: s.format(container_name=container_name))
    subprocess.run(stop_command)
    stop_command_str = " ".join(stop_command)

    while True:
        status = get_docker_container_status(container_name=container_name)
        if status is None:
            break
        logger.info(f"Waiting for container '{container_name}' to leave status '{status}'.")
        time.sleep(1)
    return stop_command_str


def get_docker_container_status(container_name: str,
                                docker_inspect_container_command_template: str =
                                "docker container inspect -f {{{{.State.Status}}}} {container_name}") -> Optional[str]:
    check_command = format_command_template(docker_inspect_container_command_template,
                                            lambda s: s.format(container_name=container_name))
    process_result = subprocess.run(check_command, stdout=subprocess.PIPE, stderr=DEVNULL)
    if process_result.returncode != 0:
        return None
    return process_result.stdout.decode("utf-8")[:-1]


def is_docker_container_running(container_name: str) -> bool:
    return get_docker_container_status(container_name=container_name) == "running"


def merge_column_values(df: Optional[pd.DataFrame], columns: Iterable[str]) -> List[str]:
    if df is None:
        return list()

    values = set()

    for column in columns:
        if column in df.columns:
            values.update(df[column].dropna())

    if None in values:
        values.remove(None)

    return list(values)


def list_files_in_directory(directory_path: str) -> List[str]:
    paths_in_dir = [os.path.join(directory_path, element) for element in os.listdir(directory_path)]
    file_paths = [path for path in paths_in_dir if os.path.isfile(path)]
    return file_paths



import os
from typing import Any, Callable, Dict, List, Optional, Set, Union

import pandas as pd
from pyteomics.mgf import MGFBase
from pyteomics.mzid import MzIdentML
from pyteomics.mzml import MzML

from mmproteo.utils import log, visualization
from mmproteo.utils.formats.mgf import read_mgf
from mmproteo.utils.formats.mz import read_mzid, read_mzml


def iter_entries(iterator: Union[MGFBase, MzIdentML, MzML], logger: log.Logger = log.DEFAULT_LOGGER) \
        -> List[Dict[str, Any]]:
    logger.debug("Iterator type = " + str(type(iterator)))
    entries = list(iterator)
    logger.debug("Length: %d" % len(entries))
    if len(entries) > 0:
        logger.debug("Example:\n" + visualization.pretty_print_json(entries[0]))
    return entries


def read_parquet(filename: str, logger: log.Logger = log.DEFAULT_LOGGER) -> pd.DataFrame:
    return pd.read_parquet(filename)


_FILE_READING_CONFIG: Dict[str, Callable[[str, log.Logger], pd.DataFrame]] = {
    "mgf": read_mgf,
    "mzid": read_mzid,
    "mzml": read_mzml,
    "parquet": read_parquet,
}


def get_readable_file_extensions() -> Set[str]:
    return set(_FILE_READING_CONFIG.keys())


def read(filename: str, filename_col: Optional[str] = "%s_filename", logger: log.Logger = log.DEFAULT_LOGGER) \
        -> pd.DataFrame:
    _, ext = separate_extension(filename=filename, extensions=get_readable_file_extensions())

    if len(ext) > 0:
        logger.debug("Started reading %s file '%s'" % (ext, filename))
        df = _FILE_READING_CONFIG[ext](filename, logger)
        logger.info("Finished reading %s file '%s'" % (ext, filename))
    else:
        raise NotImplementedError
    if filename_col is not None:
        if "%s" in filename_col:
            col = filename_col % ext
        else:
            col = filename_col
        if col in df.columns:
            logger.warning("'%s' already exists in DataFrame columns. Please choose a different column name.")
        else:
            df[col] = filename.split(os.sep)[-1]
    return df


def separate_extension(filename: str, extensions: Set[str]) -> (str, str):
    lower_filename = filename.lower()
    longest_extension = ""
    for ext in extensions:
        if lower_filename.endswith("." + ext) and len(longest_extension) < len(ext):
            longest_extension = ext

    if len(longest_extension) > 0:
        base_filename = filename[:len(filename) - len(longest_extension) - len(".")]
        return base_filename, longest_extension
    else:
        return filename, ""





#!/usr/bin/python3
import os
import sys

from mmproteo.utils import commands, log
from mmproteo.utils.config import Config

APPLICATION_NAME: str = "MMProteo"
_HERE: str = "."


def create_logger(config: Config) -> log.Logger:
    if config.dummy_logger:
        return log.DummyLogger(send_welcome=True,
                               fail_early=config.fail_early,
                               terminate_process=config.terminate_process,
                               verbose=config.verbose)

    if config.log_to_stdout:
        log_to_std = sys.stdout
    else:
        log_to_std = sys.stderr

    if config.log_file is not None and len(config.log_file) == 0:
        log_dir = None  # this disables file logging
    else:
        log_dir = config.storage_dir

    logger = log.create_logger(name=APPLICATION_NAME,
                               log_dir=log_dir,
                               filename=config.log_file,
                               verbose=config.verbose,
                               log_to_std=log_to_std,
                               fail_early=config.fail_early)
    return logger


def main(config: Config = None, logger: log.Logger = None):
    if config is None:
        config = Config()
        if logger is not None:
            config.set_logger(logger)
        config.parse_arguments()

    try:
        config.validate_arguments()

        if logger is not None:
            config.set_logger(logger)
            config.check()
        else:
            config.check()
            logger = create_logger(config)
            config.set_logger(logger)

        if config.storage_dir is not None:
            os.chdir(config.storage_dir)

        config.storage_dir = _HERE

        commands.DISPATCHER.dispatch_commands(config=config, logger=logger)
    except log.LoggedErrorException:
        return
    except KeyboardInterrupt:
        logger.info("Received KeyboardInterrupt - Shutting down")
        return


if __name__ == '__main__':
    main()



import os
import subprocess
import time
from typing import List, Optional, NoReturn

from mmproteo.utils import log, utils
from mmproteo.utils.config import Config
from mmproteo.utils.filters import AbstractFilterConditionNode, filter_files_list
from mmproteo.utils.formats.read import separate_extension
from mmproteo.utils.processing import ItemProcessor


def start_thermo_docker_container(storage_dir: str = Config.default_storage_dir,
                                  thermo_docker_container_name: str = Config.default_thermo_docker_container_name,
                                  thermo_docker_image: str = Config.default_thermo_docker_image,
                                  thermo_start_container_command_template: str =
                                  Config.default_thermo_start_container_command_template,
                                  logger: log.Logger = log.DEFAULT_LOGGER) -> None:
    subject = "ThermoRawFileParser Docker container"

    if utils.is_docker_container_running(thermo_docker_container_name):
        logger.info(subject + " is already running")
        return

    storage_dir = os.path.abspath(storage_dir)

    start_command = utils.format_command_template(command_template=thermo_start_container_command_template,
                                                  formatter=lambda s: s.format(abs_storage_dir=storage_dir,
                                                                               container_name=
                                                                               thermo_docker_container_name,
                                                                               image_name=thermo_docker_image))
    start_command_str = " ".join(start_command)
    logger.debug(f"Starting {subject} using '{start_command_str}'")
    process_result = subprocess.run(start_command)
    logger.assert_true(process_result.returncode == 0, "Failed to start " + subject)

    time.sleep(1)
    logger.assert_true(condition=utils.is_docker_container_running(thermo_docker_container_name),
                       error_msg=subject + " doesn't seem to be running")
    logger.info("Started " + subject)


def stop_thermo_docker_container(thermo_docker_container_name: str = Config.default_thermo_docker_container_name,
                                 logger: log.Logger = log.DEFAULT_LOGGER) -> None:
    subject = "ThermoRawFileParser Docker container"

    if not utils.is_docker_container_running(thermo_docker_container_name):
        logger.info(subject + " is already stopped")
        return

    logger.debug(f"Stopping {subject}")
    stop_command = utils.stop_docker_container(container_name=thermo_docker_container_name)
    logger.assert_true(condition=not utils.is_docker_container_running(thermo_docker_container_name),
                       error_msg=subject + " still seems to be running")
    logger.info(f"Stopped {subject} using {stop_command}")


_THERMO_RAW_FILE_PARSER_OUTPUT_FORMAT_IDS = {
    "mgf": 0,
    "mzml": 1,
    "imzml": 2,
    "parquet": 3,
}


def get_thermo_raw_file_parser_output_formats() -> List[str]:
    return sorted(_THERMO_RAW_FILE_PARSER_OUTPUT_FORMAT_IDS.keys())


def get_string_of_thermo_raw_file_parser_output_formats(format_quote: str = Config.default_option_quote,
                                                        separator: str = Config.default_option_separator) -> str:
    return utils.concat_set_of_options(options=get_thermo_raw_file_parser_output_formats(),
                                       option_quote=format_quote,
                                       separator=separator)


def assert_valid_thermo_output_format(output_format: str, logger: log.Logger = log.DEFAULT_LOGGER) -> \
        Optional[NoReturn]:
    logger.assert_true(output_format in get_thermo_raw_file_parser_output_formats(),
                       "Invalid output format '%s'. Currently allowed formats are: [%s]"
                       % (output_format, get_string_of_thermo_raw_file_parser_output_formats()))


def convert_raw_file(filename: Optional[str],
                     output_format: str = Config.default_thermo_output_format,
                     skip_existing: bool = Config.default_skip_existing,
                     thermo_docker_container_name: str = Config.default_thermo_docker_container_name,
                     thermo_exec_command: str = Config.default_thermo_exec_command,
                     logger: log.Logger = log.DEFAULT_LOGGER) -> Optional[str]:
    assert_valid_thermo_output_format(output_format=output_format, logger=logger)

    if filename is None:
        return None

    base_filename, file_ext = separate_extension(filename=filename,
                                                 extensions={"raw"})
    if len(file_ext) == 0:
        logger.debug("Cannot convert file '%s', unknown extension" % filename)
        return None

    converted_filename = base_filename + "." + output_format

    if skip_existing and os.path.isfile(converted_filename):
        logger.info('Skipping conversion, because "%s" already exists' % converted_filename)
        return converted_filename

    logger.assert_true(utils.is_docker_container_running(thermo_docker_container_name),
                       "There is no running ThermoRawFileParser Docker container with the name " +
                       thermo_docker_container_name)

    output_format_id = _THERMO_RAW_FILE_PARSER_OUTPUT_FORMAT_IDS[output_format]

    command = thermo_exec_command.format(container_name=thermo_docker_container_name,
                                         format=output_format_id,
                                         input=filename)

    logger.debug("Converting file using command '%s'" % command)
    return_code = os.system(command)
    if return_code == 0:
        logger.info("Converted file " + filename)
        return converted_filename
    else:
        logger.warning('Failed converting file "%s" (return code = %d)' % (filename, return_code))
        return None


class _RawFileProcessor:
    def __init__(self,
                 output_format: str = Config.default_thermo_output_format,
                 skip_existing: bool = Config.default_skip_existing,
                 thermo_docker_container_name: str = Config.default_thermo_docker_container_name,
                 thermo_exec_command: str = Config.default_thermo_exec_command,
                 logger: log.Logger = log.DEFAULT_LOGGER):
        self.output_format = output_format
        self.skip_existing = skip_existing
        self.thermo_docker_container_name = thermo_docker_container_name
        self.thermo_exec_command = thermo_exec_command
        self.logger = logger

    def __call__(self, filename: Optional[str]) -> Optional[str]:
        return convert_raw_file(filename=filename,
                                output_format=self.output_format,
                                skip_existing=self.skip_existing,
                                thermo_docker_container_name=self.thermo_docker_container_name,
                                thermo_exec_command=self.thermo_exec_command,
                                logger=self.logger)


def convert_raw_files(filenames: List[Optional[str]],
                      output_format: str = Config.default_thermo_output_format,
                      skip_existing: bool = Config.default_skip_existing,
                      column_filter: Optional[AbstractFilterConditionNode] = None,
                      max_num_files: Optional[int] = None,
                      count_failed_files: bool = Config.default_count_failed_files,
                      count_skipped_files: bool = Config.default_count_skipped_files,
                      thread_count: int = Config.default_thread_count,
                      keep_null_values: bool = Config.default_keep_null_values,
                      pre_filter_files: bool = Config.default_pre_filter_files,
                      thermo_docker_container_name: str = Config.default_thermo_docker_container_name,
                      thermo_exec_command: str = Config.default_thermo_exec_command,
                      logger: log.Logger = log.DEFAULT_LOGGER) -> List[Optional[str]]:
    assert_valid_thermo_output_format(output_format=output_format, logger=logger)

    if pre_filter_files:
        filenames = filter_files_list(filenames=filenames,
                                      file_extensions={"raw"},
                                      max_num_files=None,
                                      column_filter=column_filter,
                                      keep_null_values=keep_null_values,
                                      drop_duplicates=not keep_null_values,
                                      sort=not keep_null_values,
                                      logger=logger)

    file_processor = _RawFileProcessor(output_format=output_format,
                                       skip_existing=skip_existing,
                                       thermo_docker_container_name=thermo_docker_container_name,
                                       thermo_exec_command=thermo_exec_command,
                                       logger=logger)

    return list(ItemProcessor(items=filenames,
                              item_processor=file_processor,
                              action_name=f"raw2{output_format}-convert",
                              subject_name="raw file",
                              max_num_items=max_num_files,
                              thread_count=thread_count,
                              keep_null_values=keep_null_values,
                              count_failed_items=count_failed_files,
                              count_null_results=count_skipped_files,
                              logger=logger).process())



import json
from typing import List, Optional
from urllib.parse import quote

import pandas as pd

from mmproteo.utils import log, utils

pd.set_option('display.max_columns', None)
pd.set_option('display.max_rows', None)
pd.set_option('display.max_colwidth', None)
pd.set_option('display.width', None)


def urlencode_df_columns(files_df: pd.DataFrame, columns: List[str], inplace: bool = True) -> pd.DataFrame:
    if not inplace:
        files_df = files_df.copy()

    for col in columns:
        if col in files_df.columns:
            files_df[col] = files_df[col].map(lambda s: quote(s, safe='/:'))

    return files_df


def print_df(df: Optional[pd.DataFrame],
             max_num_files: Optional[int] = None,
             shown_columns: Optional[List[str]] = None,
             urlencode_columns: List[str] = None,
             reset_index: bool = True,
             logger: log.Logger = log.DEFAULT_LOGGER) -> None:
    if df is None or len(df) == 0:
        logger.debug("There are no entries to be shown")
        return

    df = df.copy()

    if max_num_files is not None and max_num_files != 0 and max_num_files < len(df):
        df = df[:max_num_files]
        logger.info("Showing only the top %d entries because of the max_num_files parameter" % max_num_files)
    if shown_columns is not None and shown_columns != []:
        columns = [col for col in shown_columns if col in df.columns]
        columns = utils.deduplicate_list(columns)
        df = df[columns]
        logger.info("Limiting the shown columns according to the shown_columns parameter")

    if urlencode_columns is not None:
        df = urlencode_df_columns(df, urlencode_columns, inplace=False)

    if reset_index:
        df = df.reset_index(drop=True)

    try:
        print(df)
    except BrokenPipeError:
        pass


def pretty_print_json(dic: Optional[dict]) -> str:
    if dic is None:
        return ""
    try:
        return json.dumps(dic, indent=4)
    except TypeError:
        return str(dic)



from typing import Dict, List

from mmproteo.utils import filters, log, pride, utils, visualization
from mmproteo.utils.config import Config
from mmproteo.utils.formats import archives


class AbstractCommand:
    def get_command(self) -> str:
        raise NotImplementedError

    def get_description(self) -> str:
        raise NotImplementedError

    def run(self, config: Config, logger: log.Logger = log.DEFAULT_LOGGER) -> None:
        raise NotImplementedError

    def validate(self, config: Config, logger: log.Logger = log.DEFAULT_LOGGER) -> None:
        pass


class DownloadCommand(AbstractCommand):

    def get_command(self) -> str:
        return "download"

    def get_description(self) -> str:
        return "download files from a given project."

    def run(self, config: Config, logger: log.Logger = log.DEFAULT_LOGGER) -> None:
        downloaded_files = pride.download(project_name=config.pride_project,
                                          project_files=config.get_project_files(),
                                          valid_file_extensions=config.valid_file_extensions,
                                          max_num_files=config.max_num_files,
                                          column_filter=config.column_filter,
                                          download_dir=config.storage_dir,
                                          skip_existing=config.skip_existing,
                                          count_failed_files=config.count_failed_files,
                                          count_skipped_files=config.count_skipped_files,
                                          file_name_column=config.default_file_name_column,
                                          download_link_column=config.default_download_link_column,
                                          downloaded_files_column=config.default_downloaded_files_column,
                                          api_versions=config.pride_versions,
                                          thread_count=config.thread_count,
                                          logger=logger)

        downloaded_files = downloaded_files.dropna(subset=[config.default_downloaded_files_column])

        config.cache_processed_files(data_df=downloaded_files)
        visualization.print_df(df=downloaded_files,
                               shown_columns=config.shown_columns + [config.default_downloaded_files_column],
                               urlencode_columns=[config.default_download_link_column],
                               logger=logger)

    def validate(self, config: Config, logger: log.Logger = log.DEFAULT_LOGGER) -> None:
        config.require_pride_project(logger=logger)

        if len(config.valid_file_extensions) != 0 and config.max_num_files % len(config.valid_file_extensions) != 0:
            logger.info(
                "max-num-files should be a multiple of the number of valid_file_extensions to make sure that "
                "files that belong together are also downloaded together")


class InfoCommand(AbstractCommand):

    def get_command(self) -> str:
        return "info"

    def get_description(self) -> str:
        return "request project information for a given project."

    def run(self, config: Config, logger: log.Logger = log.DEFAULT_LOGGER) -> None:
        project_info = pride.get_project_info(project_name=config.pride_project,
                                              api_versions=config.pride_versions,
                                              logger=logger)
        if project_info is None:
            return
        print(project_info)

    def validate(self, config: Config, logger: log.Logger = log.DEFAULT_LOGGER) -> None:
        config.require_pride_project(logger=logger)


class ListCommand(AbstractCommand):

    def get_command(self) -> str:
        return "list"

    def get_description(self) -> str:
        return "list files and their attributes in a given project."

    def run(self, config: Config, logger: log.Logger = log.DEFAULT_LOGGER) -> None:
        project_files = config.get_project_files()

        if project_files is None:
            return

        filtered_files = filters.filter_files_df(files_df=project_files,
                                                 file_extensions=config.valid_file_extensions,
                                                 column_filter=config.column_filter,
                                                 sort=True,
                                                 logger=logger)

        visualization.print_df(df=filtered_files,
                               max_num_files=config.max_num_files,
                               shown_columns=config.shown_columns,
                               urlencode_columns=[config.default_download_link_column],
                               logger=logger)

    def validate(self, config: Config, logger: log.Logger = log.DEFAULT_LOGGER) -> None:
        config.require_pride_project(logger=logger)


class ExtractCommand(AbstractCommand):

    def get_command(self) -> str:
        return "extract"

    def get_description(self) -> str:
        return "extract all downloaded archive files or, if none were downloaded, those in the data directory. " \
               "Currently, the following archive formats are supported: " + \
               archives.get_string_of_extractable_file_extensions()

    def run(self, config: Config, logger: log.Logger = log.DEFAULT_LOGGER) -> None:
        files = config.get_processed_files(config.default_downloaded_files_column)

        if len(files) == 0:
            files = utils.list_files_in_directory(config.storage_dir)

        extracted_files = archives.extract_files(filenames=files,
                                                 skip_existing=config.skip_existing,
                                                 max_num_files=config.max_num_files,
                                                 count_failed_files=config.count_failed_files,
                                                 count_skipped_files=config.count_skipped_files,
                                                 thread_count=config.thread_count,
                                                 column_filter=config.column_filter,
                                                 keep_null_values=False,
                                                 pre_filter_files=True,
                                                 logger=logger)

        result_df = config.cache_processed_files(data_list=extracted_files,
                                                 column_names=config.default_extracted_files_column)
        visualization.print_df(df=result_df, logger=logger)


class ConvertRawCommand(AbstractCommand):

    def get_command(self) -> str:
        return "convertraw"

    def get_description(self) -> str:
        return "convert all downloaded or extracted raw files or, if none were downloaded or extracted, " \
               "those raw files in the data " \
               "directory, into the given thermo output format using the ThermoRawFileParser. " \
               "This command requires an accessible Docker installation."

    def run(self, config: Config, logger: log.Logger = log.DEFAULT_LOGGER) -> None:
        files = config.get_processed_files(config.default_downloaded_files_column,
                                           config.default_extracted_files_column)

        if len(files) == 0:
            files = utils.list_files_in_directory(config.storage_dir)

        from mmproteo.utils.formats import raw
        raw.start_thermo_docker_container(storage_dir=config.storage_dir,
                                          thermo_docker_container_name=Config.default_thermo_docker_container_name,
                                          thermo_docker_image=Config.default_thermo_docker_image,
                                          thermo_start_container_command_template=Config.
                                          default_thermo_start_container_command_template,
                                          logger=logger)

        converted_files = raw.convert_raw_files(filenames=files,
                                                output_format=config.thermo_output_format,
                                                skip_existing=config.skip_existing,
                                                max_num_files=config.max_num_files,
                                                count_failed_files=config.count_failed_files,
                                                count_skipped_files=config.count_skipped_files,
                                                thread_count=config.thread_count,
                                                column_filter=config.column_filter,
                                                keep_null_values=False,
                                                pre_filter_files=True,
                                                thermo_docker_container_name=Config.
                                                default_thermo_docker_container_name,
                                                thermo_exec_command=Config.default_thermo_exec_command,
                                                logger=logger)

        result_df = config.cache_processed_files(converted_files, config.default_converted_raw_files_column)

        if not config.thermo_keep_container_running:
            raw.stop_thermo_docker_container(
                thermo_docker_container_name=Config.default_thermo_docker_container_name,
                logger=logger
            )

        visualization.print_df(df=result_df, logger=logger)

    def validate(self, config: Config, logger: log.Logger = log.DEFAULT_LOGGER) -> None:
        from mmproteo.utils.formats import raw
        raw.assert_valid_thermo_output_format(output_format=config.thermo_output_format,
                                              logger=logger)


class Mgf2ParquetCommand(AbstractCommand):

    def get_command(self) -> str:
        return "mgf2parquet"

    def get_description(self) -> str:
        return "convert all downloaded, extracted, or converted mgf files into parquet format, or, " \
               "if no files were previously processed, convert the mgf files in the data directory."

    def run(self, config: Config, logger: log.Logger = log.DEFAULT_LOGGER) -> None:
        files = config.get_processed_files(config.default_downloaded_files_column,
                                           config.default_extracted_files_column,
                                           config.default_converted_raw_files_column)

        if len(files) == 0:
            files = utils.list_files_in_directory(config.storage_dir)

        from mmproteo.utils.formats import mgf
        mgf_parquet_files = mgf.convert_mgf_files_to_parquet(filenames=files,
                                                             skip_existing=config.skip_existing,
                                                             max_num_files=config.max_num_files,
                                                             count_failed_files=config.count_failed_files,
                                                             count_skipped_files=config.count_skipped_files,
                                                             thread_count=config.thread_count,
                                                             column_filter=config.column_filter,
                                                             keep_null_values=False,
                                                             pre_filter_files=True,
                                                             logger=logger)

        result_df = config.cache_processed_files(mgf_parquet_files, config.default_mgf_parquet_files_column)
        visualization.print_df(df=result_df, logger=logger)


class Mz2ParquetCommand(AbstractCommand):
    def get_command(self) -> str:
        return "mz2parquet"

    def get_description(self) -> str:
        return "merge and convert all downloaded or extracted mzid and mzml files into parquet format" \
               " or, if no files were previously processed, merge and convert the files in the data directory."

    def run(self, config: Config, logger: log.Logger = log.DEFAULT_LOGGER) -> None:
        files = config.get_processed_files(config.default_downloaded_files_column,
                                           config.default_extracted_files_column)

        if len(files) == 0:
            files = utils.list_files_in_directory(config.storage_dir)

        from mmproteo.utils.formats import mz
        mzmlid_parquet_files = mz.merge_mzml_and_mzid_files_to_parquet(filenames=files,
                                                                       skip_existing=config.skip_existing,
                                                                       max_num_files=config.max_num_files,
                                                                       count_failed_files=config.count_failed_files,
                                                                       count_skipped_files=config.count_skipped_files,
                                                                       thread_count=config.thread_count,
                                                                       column_filter=config.column_filter,
                                                                       logger=logger)

        result_df = config.cache_processed_files(mzmlid_parquet_files, config.default_mzmlid_parquet_files_column)
        visualization.print_df(df=result_df, logger=logger)


class ShowConfigCommand(AbstractCommand):
    def get_command(self) -> str:
        return "showconfig"

    def get_description(self) -> str:
        return "print all variables of the current run configuration."

    def run(self, config: Config, logger: log.Logger = log.DEFAULT_LOGGER) -> None:
        print(config)


class CommandDispatcher:
    def __init__(self) -> None:
        self.__commands: Dict[str, AbstractCommand] = dict()

    def register(self, command: AbstractCommand) -> None:
        if command.get_command() in self.__commands.keys():
            raise ValueError("Command is already registered")
        self.__commands[command.get_command()] = command

    def get_command_names(self) -> List[str]:
        return sorted(self.__commands.keys())

    @staticmethod
    def _pad_command(command: str, width: int) -> str:
        return command.ljust(width)

    def get_command_descriptions_str(self) -> str:
        longest_command_length = max([len(command) for command in self.get_command_names()])

        return '\n'.join(
            [self._pad_command(command.get_command(), longest_command_length) + " : " + command.get_description()
             for command in self.__commands.values()])

    def get_command(self, command_name: str) -> AbstractCommand:
        if command_name not in self.__commands.keys():
            raise NotImplementedError("%s is no known command")
        return self.__commands[command_name]

    def dispatch_commands(self,
                          config: Config,
                          catch_validation_warnings: bool = True,
                          catch_run_warnings: bool = True,
                          logger: log.Logger = log.DEFAULT_LOGGER) -> None:
        commands = [self.get_command(command_name) for command_name in config.commands]

        for command in commands:
            if catch_validation_warnings:
                try:
                    command.validate(config=config, logger=logger)
                except log.LoggedWarningException:
                    pass
            else:
                command.validate(config=config, logger=logger)

        for command in commands:
            if catch_run_warnings:
                try:
                    command.run(config=config, logger=logger)
                except log.LoggedWarningException:
                    pass
            else:
                command.run(config=config, logger=logger)


DISPATCHER = CommandDispatcher()
DISPATCHER.register(ConvertRawCommand())
DISPATCHER.register(DownloadCommand())
DISPATCHER.register(ExtractCommand())
DISPATCHER.register(InfoCommand())
DISPATCHER.register(ListCommand())
DISPATCHER.register(Mgf2ParquetCommand())
DISPATCHER.register(Mz2ParquetCommand())
DISPATCHER.register(ShowConfigCommand())



from mmproteo.utils import utils


def test_flatten_single_element_containers():
    a = dict()
    assert utils.flatten_single_element_containers(a) == a
    assert utils.flatten_single_element_containers([a]) == a

    b = [1, 2]
    assert utils.flatten_single_element_containers(b) == b
    assert utils.flatten_single_element_containers([[b]]) == b

    c = "c"
    assert utils.flatten_single_element_containers(c) == c
    assert utils.flatten_single_element_containers([{c}]) == c


def test_flatten_dict_without_concatenation():
    a = {3: 4, 4: [4], 5: {5}}
    b = {0: 0, 1: [1], 2: {2}, 3: 3, "a": a}

    res = {0: 0, 1: 1, 2: 2, 3: 3, 4: 4, 5: 5}
    assert utils.flatten_dict(b, concat_keys=False, clean_keys=False) == res


def test_flatten_dict_with_concatenation():
    a = {3: 4, 4: [4], 5: {5}}
    b = {0: 0, 1: [1], 2: {2}, 3: 3, "a": a}

    res = {"0": 0, "1": 1, "2": 2, "3": 3, "a__3": 4, "a__4": 4, "a__5": 5}
    assert utils.flatten_dict(b, concat_keys=True, clean_keys=False) == res



import os

DEFAULT_PROJECT = "PXD010000"
DEFAULT_PROJECT_FILE_COUNT = 1175
DEFAULT_TEST_API = "2"

DATASET_PATH = os.path.join("..", "..", "datasets")
DEFAULT_PROJECT_DATASET_PATH = os.path.join(DATASET_PATH, DEFAULT_PROJECT)
os.makedirs(DEFAULT_PROJECT_DATASET_PATH, exist_ok=True)

MZML_FILE = "Biodiversity_A_cryptum_FeTSB_anaerobic_1_01Jun16_Pippin_16-03-39.mzML"
MZML_GZ_FILE = MZML_FILE + ".gz"
MZML_FILE_PATH = os.path.join(DEFAULT_PROJECT_DATASET_PATH, MZML_FILE)

MZID_FILE = "Biodiversity_A_cryptum_FeTSB_anaerobic_1_01Jun16_Pippin_16-03-39_msgfplus.mzid"
MZID_GZ_FILE = MZID_FILE + ".gz"
MZID_FILE_PATH = os.path.join(DEFAULT_PROJECT_DATASET_PATH, MZID_FILE)

MZMLID_FILE = "Biodiversity_A_cryptum_FeTSB_anaerobic_1_01Jun16_Pippin_16-03-39_mzmlid.parquet"
MZMLID_FILE_PATH = os.path.join(DEFAULT_PROJECT_DATASET_PATH, MZMLID_FILE)

RAW_FILE = "Biodiversity_A_cryptum_FeTSB_anaerobic_1_01Jun16_Pippin_16-03-39.raw"
MGF_FILE = "Biodiversity_A_cryptum_FeTSB_anaerobic_1_01Jun16_Pippin_16-03-39.mgf"
MGF_PARQUET_FILE = "Biodiversity_A_cryptum_FeTSB_anaerobic_1_01Jun16_Pippin_16-03-39.parquet"




#!/usr/bin/python3

import subprocess
import time

DEFAULT_PROJECT = "PXD010000"
FAIR_USE_DELAY_SECONDS = 0.5


def _store_command_output(command: str, filename: str, include_stderr: bool = True) -> None:
    system_command = f"{command} > {filename}"
    if include_stderr:
        system_command += " 2>&1"
    subprocess.run(system_command, shell=True)


def _compare_stdout_with_file(command: str, filename: str, include_stderr: bool = True) -> None:
    if include_stderr:
        command += " 2>&1"

    process = subprocess.run(command, shell=True, stdout=subprocess.PIPE)
    stdout = process.stdout.decode('utf-8')
    stdout_lines = stdout.split("\n")

    stdout_lines = stdout_lines[:-1]  # remove empty trailing line

    with open(filename) as file:
        file_lines = file.readlines()
    file_lines = [line.rstrip("\n") for line in file_lines]

    assert len(stdout_lines) == len(file_lines), "outputs do not have the same lengths"
    for received, expected in zip(stdout_lines, file_lines):
        assert received == expected, "found lines that do not match"


def store_mmproteo_output():
    _store_command_output(command="mmproteo",
                          filename="resources/mmproteo_output.txt")


def test_mmproteo_output():
    _compare_stdout_with_file(command="mmproteo",
                              filename="resources/mmproteo_output.txt")


def store_mmproteo_h_output():
    _store_command_output(command="mmproteo -h",
                          filename="resources/mmproteo_h_output.txt")


def test_mmproteo_h_output():
    _compare_stdout_with_file(command="mmproteo -h",
                              filename="resources/mmproteo_h_output.txt")


def store_mmproteo_p_info_output():
    _store_command_output(command=f"mmproteo -p {DEFAULT_PROJECT} info",
                          filename="resources/mmproteo_p_info_output.txt",
                          include_stderr=False)
    time.sleep(FAIR_USE_DELAY_SECONDS)


def test_mmproteo_p_info_output():
    _compare_stdout_with_file(command=f"mmproteo -p {DEFAULT_PROJECT} info",
                              filename="resources/mmproteo_p_info_output.txt",
                              include_stderr=False)
    time.sleep(FAIR_USE_DELAY_SECONDS)


def store_mmproteo_p_list_output():
    _store_command_output(command=f"mmproteo -p {DEFAULT_PROJECT} list",
                          filename="resources/mmproteo_p_list_output.txt",
                          include_stderr=False)
    time.sleep(FAIR_USE_DELAY_SECONDS)


def test_mmproteo_p_list_output():
    _compare_stdout_with_file(command=f"mmproteo -p {DEFAULT_PROJECT} list",
                              filename="resources/mmproteo_p_list_output.txt",
                              include_stderr=False)
    time.sleep(FAIR_USE_DELAY_SECONDS)


def store_mmproteo_p_n_list_output():
    _store_command_output(command=f"mmproteo -p {DEFAULT_PROJECT} -n 10 list",
                          filename="resources/mmproteo_p_n_list_output.txt",
                          include_stderr=False)
    time.sleep(FAIR_USE_DELAY_SECONDS)


def test_mmproteo_p_n_list_output():
    _compare_stdout_with_file(command=f"mmproteo -p {DEFAULT_PROJECT} -n 10 list",
                              filename="resources/mmproteo_p_n_list_output.txt",
                              include_stderr=False)
    time.sleep(FAIR_USE_DELAY_SECONDS)


def store_mmproteo_p_e_mzid_list_output():
    _store_command_output(command=f"mmproteo -p {DEFAULT_PROJECT} -e mzid list",
                          filename="resources/mmproteo_p_e_mzid_list_output.txt",
                          include_stderr=False)
    time.sleep(FAIR_USE_DELAY_SECONDS)


def test_mmproteo_p_e_mzid_list_output():
    _compare_stdout_with_file(command=f"mmproteo -p {DEFAULT_PROJECT} -e mzid list",
                              filename="resources/mmproteo_p_e_mzid_list_output.txt",
                              include_stderr=False)
    time.sleep(FAIR_USE_DELAY_SECONDS)


def store_mmproteo_p_e_gz_list_output():
    _store_command_output(command=f"mmproteo -p {DEFAULT_PROJECT} -e gz list",
                          filename="resources/mmproteo_p_e_gz_list_output.txt",
                          include_stderr=False)
    time.sleep(FAIR_USE_DELAY_SECONDS)


def test_mmproteo_p_e_gz_list_output():
    _compare_stdout_with_file(command=f"mmproteo -p {DEFAULT_PROJECT} -e gz list",
                              filename="resources/mmproteo_p_e_gz_list_output.txt",
                              include_stderr=False)
    time.sleep(FAIR_USE_DELAY_SECONDS)


def store_mmproteo_p_e_gz_mzid_list_output():
    _store_command_output(command=f"mmproteo -p {DEFAULT_PROJECT} -e gz,mzid list",
                          filename="resources/mmproteo_p_e_gz_mzid_list_output.txt",
                          include_stderr=False)
    time.sleep(FAIR_USE_DELAY_SECONDS)


def test_mmproteo_p_e_gz_mzid_list_output():
    _compare_stdout_with_file(command=f"mmproteo -p {DEFAULT_PROJECT} -e gz,mzid list",
                              filename="resources/mmproteo_p_e_gz_mzid_list_output.txt",
                              include_stderr=False)
    time.sleep(FAIR_USE_DELAY_SECONDS)


def store_mmproteo_p_c_list_output():
    _store_command_output(command=f"mmproteo -p {DEFAULT_PROJECT} -c fileName,downloadLink list",
                          filename="resources/mmproteo_p_c_list_output.txt",
                          include_stderr=False)
    time.sleep(FAIR_USE_DELAY_SECONDS)


def test_mmproteo_p_c_list_output():
    _compare_stdout_with_file(command=f"mmproteo -p {DEFAULT_PROJECT} -c fileName,downloadLink list",
                              filename="resources/mmproteo_p_c_list_output.txt",
                              include_stderr=False)
    time.sleep(FAIR_USE_DELAY_SECONDS)


def store_mmproteo_p_n_e_c_list_output():
    _store_command_output(command=f"mmproteo -p {DEFAULT_PROJECT} -n 100 -e raw,mzml,gz -c fileName,downloadLink list",
                          filename="resources/mmproteo_p_n_e_c_list_output.txt",
                          include_stderr=False)
    time.sleep(FAIR_USE_DELAY_SECONDS)


def test_mmproteo_p_n_e_c_list_output():
    _compare_stdout_with_file(command=f"mmproteo -p {DEFAULT_PROJECT} -n 100 -e raw,mzml,gz "
                                      f"-c fileName,downloadLink list",
                              filename="resources/mmproteo_p_n_e_c_list_output.txt",
                              include_stderr=False)
    time.sleep(FAIR_USE_DELAY_SECONDS)


def store_mmproteo_showconfig_output():
    _store_command_output(command="mmproteo showconfig",
                          filename="resources/mmproteo_showconfig_output.txt",
                          include_stderr=False)


def test_mmproteo_showconfig_output():
    _compare_stdout_with_file(command="mmproteo showconfig",
                              filename="resources/mmproteo_showconfig_output.txt",
                              include_stderr=False)


if __name__ == '__main__':
    # run this file as script to recreate the expected outputs
    store_mmproteo_output()
    store_mmproteo_h_output()
    store_mmproteo_p_info_output()
    store_mmproteo_p_list_output()
    store_mmproteo_p_n_list_output()
    store_mmproteo_p_e_mzid_list_output()
    store_mmproteo_p_e_gz_list_output()
    store_mmproteo_p_e_gz_mzid_list_output()
    store_mmproteo_p_c_list_output()
    store_mmproteo_p_n_e_c_list_output()
    store_mmproteo_showconfig_output()



import os
from typing import Optional, List

import pandas as pd
from pyteomics import mgf

from mmproteo.utils import log, utils
from mmproteo.utils.config import Config
from mmproteo.utils.filters import AbstractFilterConditionNode, filter_files_list
from mmproteo.utils.formats import read
from mmproteo.utils.processing import ItemProcessor


def read_mgf(filename: str, logger: log.Logger = log.DEFAULT_LOGGER) -> pd.DataFrame:
    entries = read.iter_entries(mgf.read(filename), logger=logger)
    extracted_entries = [utils.flatten_dict(entry) for entry in entries]
    return pd.DataFrame(data=extracted_entries)


def convert_mgf_file_to_parquet(filename: Optional[str],
                                skip_existing: bool = Config.default_skip_existing,
                                logger: log.Logger = log.DEFAULT_LOGGER) -> Optional[str]:
    if filename is None:
        return None

    base_filename, file_ext = read.separate_extension(filename=filename,
                                                      extensions={"mgf"})
    if len(file_ext) == 0:
        logger.debug("Cannot convert file '%s', unknown extension" % filename)
        return None

    converted_filename = base_filename + ".parquet"

    if skip_existing and os.path.isfile(converted_filename):
        logger.info('Skipping conversion, because "%s" already exists' % converted_filename)
        return converted_filename

    try:
        df = read_mgf(filename=filename, logger=logger)
        df.to_parquet(path=converted_filename)
        logger.info("Converted file " + filename)
        return converted_filename
    except Exception as e:
        logger.warning(f'Failed converting file "{filename}" ({e})')
        return None


class _Mgf2ParquetFileProcessor:
    def __init__(self,
                 skip_existing: bool = Config.default_skip_existing,
                 logger: log.Logger = log.DEFAULT_LOGGER):
        self.skip_existing = skip_existing
        self.logger = logger

    def __call__(self, filename: Optional[str]) -> Optional[str]:
        return convert_mgf_file_to_parquet(filename=filename,
                                           skip_existing=self.skip_existing,
                                           logger=self.logger)


def convert_mgf_files_to_parquet(filenames: List[Optional[str]],
                                 skip_existing: bool = Config.default_skip_existing,
                                 max_num_files: Optional[int] = None,
                                 count_failed_files: bool = Config.default_count_failed_files,
                                 count_skipped_files: bool = Config.default_count_skipped_files,
                                 thread_count: int = Config.default_thread_count,
                                 column_filter: Optional[AbstractFilterConditionNode] = None,
                                 keep_null_values: bool = Config.default_keep_null_values,
                                 pre_filter_files: bool = Config.default_pre_filter_files,
                                 logger: log.Logger = log.DEFAULT_LOGGER) -> List[Optional[str]]:
    if pre_filter_files:
        filenames = filter_files_list(filenames=filenames,
                                      file_extensions={"mgf"},
                                      max_num_files=None,
                                      column_filter=column_filter,
                                      keep_null_values=keep_null_values,
                                      drop_duplicates=not keep_null_values,
                                      sort=not keep_null_values,
                                      logger=logger)

    file_processor = _Mgf2ParquetFileProcessor(skip_existing=skip_existing, logger=logger)

    return list(ItemProcessor(items=filenames,
                              item_processor=file_processor,
                              action_name="mgf2parquet-convert",
                              subject_name="MGF file",
                              max_num_items=max_num_files,
                              count_failed_items=count_failed_files,
                              count_null_results=count_skipped_files,
                              thread_count=thread_count,
                              keep_null_values=keep_null_values,
                              logger=logger).process())



import tempfile
import shutil
import os

import pytest

from mmproteo.utils.config import Config
from .defaults import DEFAULT_PROJECT_DATASET_PATH, MZML_FILE_PATH, MZID_FILE_PATH


@pytest.fixture(scope="function")
def run_in_temp_directory():
    # before the test
    current_working_dir = os.getcwd()
    temporary_working_dir = tempfile.mkdtemp()
    os.chdir(temporary_working_dir)
    print(f"working in '{temporary_working_dir}'")

    yield  # run the test

    # after the test
    os.chdir(current_working_dir)
    shutil.rmtree(temporary_working_dir)
    print("removed temporary working directory")


@pytest.fixture(scope="session")
def run_with_datasets():
    if os.path.isfile(MZML_FILE_PATH) and os.path.isfile(MZID_FILE_PATH):
        return

    current_working_dir = os.getcwd()
    os.chdir(DEFAULT_PROJECT_DATASET_PATH)
    os.system("mmproteo -p PXD010000 -e mzid,mzml -n 2 download extract")
    os.remove(Config.default_log_file)
    os.chdir(current_working_dir)



import json
import os
from typing import List, NoReturn, Optional, Union, Tuple

import requests
import wget
from requests import Response

from mmproteo.utils import log, utils
from mmproteo.utils.config import Config
from mmproteo.utils.formats import archives, read
from mmproteo.utils.processing import ItemProcessor
from mmproteo.utils.visualization import pretty_print_json


def download_file(download_url: str, skip_existing: bool = Config.default_skip_existing) -> (str, Optional[str]):
    filename = download_url.split("/")[-1]
    downloaded_file_name = None

    found_downloaded_file = False
    found_extracted_file = False

    skip_reason = None

    if len(filename) > 0 and skip_existing:
        if os.path.isfile(filename):
            found_downloaded_file = True
            downloaded_file_name = filename
            skip_reason = 'file "%s" already exists' % downloaded_file_name

        extracted_file_name, extension = read.separate_extension(filename, archives.get_extractable_file_extensions())
        file_is_extractable = len(extension) > 0

        if file_is_extractable:
            # gunzip deletes the archive file after extraction, other programmes keep it
            if os.path.isfile(extracted_file_name):
                found_extracted_file = True
                skip_reason = 'extracted file "%s" already exists' % extracted_file_name

    if not (skip_existing and (found_downloaded_file or found_extracted_file)):
        downloaded_file_name = wget.download(download_url)  # might raise an EXCEPTION

    return downloaded_file_name, skip_reason


class _DownloadUrlProcessor:
    def __init__(self,
                 download_url_count: int,
                 skip_existing: bool = Config.default_skip_existing,
                 logger: log.Logger = log.DEFAULT_LOGGER):
        self.download_url_count = download_url_count
        self.skip_existing = skip_existing
        self.logger = logger

    def __call__(self, indexed_url: Tuple[int, str]) -> Union[Optional[str], NoReturn]:
        current_download_index, url = indexed_url
        return handle_file_download(download_url=url,
                                    current_download_index=current_download_index,
                                    download_count=self.download_url_count,
                                    skip_existing=self.skip_existing,
                                    logger=self.logger)


def handle_file_download(download_url: str,
                         current_download_index: int,
                         download_count: int,
                         skip_existing: bool = Config.default_skip_existing,
                         logger: log.Logger = log.DEFAULT_LOGGER) -> Union[Optional[str], NoReturn]:
    logger.info(f"Downloading file {current_download_index + 1}/{download_count}: {download_url}")

    try:
        downloaded_file_name, skip_reason = download_file(download_url, skip_existing)
    except Exception as e:
        logger.info(f'Failed to download file {current_download_index + 1}/{download_count} ("{download_url}") '
                    f'because of "{e}"')
        raise

    if skip_reason is not None:
        logger.info('Skipped download, because ' + skip_reason)
        return None
    else:
        logger.info(f'Downloaded file {current_download_index + 1}/{download_count}: "{download_url}"')
        return downloaded_file_name


def download_files(download_urls: List[str],
                   skip_existing: bool = Config.default_skip_existing,
                   max_num_files: Optional[int] = None,
                   count_skipped_files: bool = Config.default_count_skipped_files,
                   count_failed_files: bool = Config.default_count_failed_files,
                   keep_null_values: bool = Config.default_keep_null_values,
                   thread_count: int = Config.default_thread_count,
                   logger: log.Logger = log.DEFAULT_LOGGER) -> List[Optional[str]]:
    download_count = len(download_urls)
    logger.info(f"Downloading {download_count} file{utils.get_plural_s(download_count)}")

    download_url_processor = _DownloadUrlProcessor(download_url_count=download_count,
                                                   skip_existing=skip_existing,
                                                   logger=logger)
    item_processor = ItemProcessor(items=enumerate(download_urls),
                                   item_processor=download_url_processor,
                                   action_name="download",
                                   subject_name="URL",
                                   keep_null_values=keep_null_values,
                                   max_num_items=max_num_files,
                                   count_null_results=count_skipped_files,
                                   count_failed_items=count_failed_files,
                                   thread_count=thread_count,
                                   logger=logger)
    downloaded_files_names: List[Optional[str]] = list(item_processor.process())
    return downloaded_files_names


class AbstractDownloader:
    def __init__(self, logger: log.Logger = log.DEFAULT_LOGGER):
        self.logger = logger

    @staticmethod
    # HTTP 204 - No Content
    def _handle_204_response(response_dict: dict, logger: log.Logger = log.DEFAULT_LOGGER) -> Optional[NoReturn]:
        logger.warning("Repository does not exist")

    @staticmethod
    # HTTP 401 - Unauthorized
    def _handle_401_response(response_dict: dict, logger: log.Logger = log.DEFAULT_LOGGER) -> Optional[NoReturn]:
        message = response_dict.get('message', "?")
        developer_message = response_dict.get('developerMessage', "?")
        more_info_url = response_dict.get('moreInfoUrl', "?")
        logger.warning("%s (%s) -> %s" % (message, developer_message, more_info_url))

    @staticmethod
    def _handle_unknown_response(status_code: int, response_dict: dict,
                                 logger: log.Logger = log.DEFAULT_LOGGER) -> Optional[NoReturn]:
        logger.warning("Received unknown response code %d or content" % status_code)
        logger.debug(pretty_print_json(response_dict))

    def _handle_non_200_response_codes(self, response: Optional[Response],
                                       logger: log.Logger = log.DEFAULT_LOGGER) -> Optional[NoReturn]:
        if response is None:
            return None
        if response.status_code == 204:
            return self._handle_204_response(logger=logger)
        try:
            response_dict = json.loads(response.text)
        except json.JSONDecodeError:
            logger.warning("Received unknown non-JSON response with response code %d" % response.status_code)
            logger.debug("Response text: '%s'" % response.text)
            return
        if response.status_code == 401:
            return self._handle_401_response(response_dict, logger)
        return self._handle_unknown_response(response.status_code, response_dict, logger)

    def request_json(self, url: str, subject_name: str, logger: log.Logger = log.DEFAULT_LOGGER) -> Union[
        Optional[dict],
        Optional[NoReturn],
    ]:
        logger.info("Requesting %s from %s" % (subject_name, url))
        response = requests.get(url)
        logger.debug("Received response from %s with length of %d bytes and status code %d" %
                     (url, len(response.text), response.status_code))

        if response.status_code == 200:
            response_dict = json.loads(response.text)
            return response_dict
        else:
            self._handle_non_200_response_codes(response, logger)
            return None



import os

from mmproteo.utils import filters
from mmproteo.utils.formats import mz, read
from .utils import defaults
from .utils.fixtures import run_with_datasets


def test_read_mzid(run_with_datasets):
    read.read(defaults.MZID_FILE_PATH)


def test_merge_mzml_mzid_to_parquet(run_with_datasets):
    if os.path.isfile(defaults.MZMLID_FILE_PATH):
        os.remove(defaults.MZMLID_FILE_PATH)
    parquet_files = mz.merge_mzml_and_mzid_files_to_parquet(filenames=[defaults.MZML_FILE_PATH,
                                                                       defaults.MZID_FILE_PATH])
    assert os.path.isfile(defaults.MZMLID_FILE_PATH)
    assert parquet_files == [defaults.MZMLID_FILE_PATH]
    os.remove(defaults.MZMLID_FILE_PATH)


def test_filter_files_list():
    filenames = [
        "abc.txt",
        "def.txt.gz",
        "ghi.gz"
    ]

    assert ["abc.txt"] == filters.filter_files_list(filenames=filenames, max_num_files=1)
    assert ["abc.txt", "def.txt.gz"] == filters.filter_files_list(filenames=filenames, file_extensions=["txt"])
    assert ["def.txt.gz", "ghi.gz"] == filters.filter_files_list(filenames=filenames, file_extensions=["gz"])
    assert ["abc.txt", "def.txt.gz", "ghi.gz"] == filters.filter_files_list(filenames=filenames,
                                                                            file_extensions=["gz", "txt"])



import logging
import os
import sys
from typing import List, NoReturn, Optional, TextIO

LOG_FORMAT: str = '%(asctime)s - %(name)s: %(message)s'
DEFAULT_LOG_SUFFIX: str = '.log'


class LoggedWarningException(Exception):
    def __init__(self, msg: str):
        super().__init__(msg)


class LoggedErrorException(Exception):
    def __init__(self, msg: str):
        super().__init__(msg)


class Logger:
    def __init__(self,
                 logger: Optional[logging.Logger],
                 fail_early: bool = True,
                 terminate_process: bool = False,
                 verbose: Optional[bool] = None):
        assert logger is None or verbose is None, "cannot use 'logger' and 'verbose' parameter simultaneously. " \
                                                  "'verbose' is only considered if 'logger' is None"
        assert logger is not None or verbose is not None, "either 'logger' or 'verbose' must be given to define " \
                                                          "the log level"
        self.logger = logger
        self.fail_early = fail_early
        self.terminate_process = terminate_process
        self._verbose = verbose

    def __print(self, msg: str):
        print(msg)

    def info(self, msg: str = "") -> None:
        if self.logger is not None:
            self.logger.info(msg)
        else:
            self.__print("INFO: " + msg)

    def debug(self, msg: str = "") -> None:
        if self.logger is not None:
            self.logger.debug("DEBUG: " + msg)
        elif self._verbose:
            self.__print("DEBUG: " + msg)

    def error(self, msg: str = "") -> NoReturn:
        if self.logger is not None:
            self.logger.info("ERROR: " + msg)
        else:
            self.__print("ERROR: " + msg)
        if self.terminate_process:
            sys.exit(1)
        raise LoggedErrorException(msg)

    def warning(self, msg: str = "") -> Optional[NoReturn]:
        if self.logger is not None:
            self.logger.info("WARNING: " + msg)
        else:
            self.__print("WARNING: " + msg)
        if self.fail_early:
            if self.terminate_process:
                self.info("Shutting down because of fail-early configuration")
                sys.exit(1)
            raise LoggedWarningException(msg)

    def assert_true(self, condition: bool, error_msg: str) -> Optional[NoReturn]:
        if not condition:
            self.error(error_msg)

    def is_verbose(self) -> bool:
        if self.logger is not None:
            return self.logger.getEffectiveLevel() == logging.DEBUG
        else:
            return self._verbose


class DummyLogger(Logger):
    def __init__(self,
                 send_welcome: bool = True,
                 fail_early: bool = False,
                 terminate_process: bool = False,
                 verbose: bool = False):
        super().__init__(logger=None,
                         fail_early=fail_early,
                         terminate_process=terminate_process,
                         verbose=verbose)
        if send_welcome:
            self.info("Printing to Stdout")


class TestLogger(Logger):
    def __init__(self,
                 fail_early: bool = False,
                 terminate_process: bool = False,
                 verbose: bool = False):
        super().__init__(logger=None,
                         fail_early=fail_early,
                         terminate_process=terminate_process,
                         verbose=verbose)
        self.msg_buffer: List[str] = []

    def __print(self, msg: str):
        self.msg_buffer.append(msg)


def create_logger(name: str,
                  level: Optional[int] = None,
                  verbose: Optional[bool] = None,
                  filename: Optional[str] = None,
                  log_dir: Optional[str] = ".",
                  log_to_std: Optional[TextIO] = sys.stderr,
                  fail_early: bool = False) -> Logger:
    """

    :param name:
    :param level:
    :param verbose:
    :param filename:    the name of the file in the :param:log_dir to which the log is written.
                        If it is None, the log filename is constructed as ':param:`name`.log'.
    :param log_dir:     the path of the directory in which the log file is created.
                        If it is None, file logging is disabled.
    :param log_to_std:  an IO handler (e.g. :attr:`sys.stderr` or :attr:`sys.stdout`), usually used for command line
                        log output. If it is None, command line logging is disabled.
    :param fail_early:
    :return:
    """
    if verbose is not None:
        assert level is None, "level and verbose are exclusive parameters"
        if verbose:
            level = logging.DEBUG
        else:
            level = logging.INFO
    else:
        assert level is not None, "either level or verbose must be given as parameter"

    assert name is not None, "the 'name' parameter is required as logger name"
    assert log_dir is not None or log_to_std is not None, \
        "either 'log_dir' or 'log_to_std' must be given to specify a log target"

    try:
        if filename is None:
            filename = name + DEFAULT_LOG_SUFFIX
        if log_dir is not None:
            filename = os.path.join(log_dir, filename)
        logger = logging.getLogger(name)
        logger.setLevel(level)

        formatter = logging.Formatter(LOG_FORMAT)

        target_names = list()

        if log_dir is not None:
            # log to file
            dir_name = os.path.dirname(filename)

            from mmproteo.utils.utils import ensure_dir_exists  # prevent circular import
            ensure_dir_exists(dir_name)
            file_logging_handler = logging.FileHandler(filename)
            file_logging_handler.setFormatter(formatter)
            file_logging_handler.setLevel(level)
            logger.addHandler(file_logging_handler)

            target_names.append(f"file '{filename}'")

        if log_to_std is not None:
            std_logging_handler = logging.StreamHandler(log_to_std)
            std_logging_handler.flush = sys.stdout.flush
            std_logging_handler.setFormatter(formatter)
            std_logging_handler.setLevel(level)
            logger.addHandler(std_logging_handler)

            target_names.append(log_to_std.name)

        logger.info("Logging to " + " and to ".join([target for target in target_names]))

        return Logger(logger, fail_early=fail_early)
    except Exception as e:
        logger = DummyLogger(send_welcome=True, fail_early=fail_early)
        logger.debug(str(e))
        logger.warning("Failed to create logger " + str(name))
        logger.info("Returned print-only dummy logger")

        return logger


DEFAULT_LOGGER: Logger = DummyLogger(send_welcome=False)



import gc
import multiprocessing
import signal
from typing import Callable, Any, Optional, Tuple, Iterable, List, NoReturn, Union

from mmproteo.utils import log, utils
from mmproteo.utils.config import Config


class _IndexedItemProcessor:
    def __init__(self, item_processor: Callable[[Any], Union[Optional[str], NoReturn]]):
        self.item_processor = item_processor

    def __call__(self, indexed_item: Tuple[int, Optional[Any]]) -> Optional[Any]:
        index, item = indexed_item
        if item is None:
            response = None
        else:
            try:
                response = self.item_processor(item)
            except KeyboardInterrupt:
                raise
            except Exception as e:
                response = e
        gc.collect()
        return index, response


class ItemProcessor:
    def __init__(self,
                 items: Iterable[Optional[Any]],
                 item_processor: Callable[[Any], Union[Optional[str], NoReturn]],
                 action_name: str,
                 action_name_past_form: Optional[str] = None,
                 subject_name: str = "file",
                 max_num_items: Optional[int] = None,
                 keep_null_values: bool = Config.default_keep_null_values,
                 count_null_results: bool = Config.default_count_null_results,
                 count_failed_items: bool = Config.default_count_failed_files,
                 thread_count: int = Config.default_thread_count,
                 logger: log.Logger = log.DEFAULT_LOGGER):

        if max_num_items == 0:
            max_num_items = None

        self.items = list(items)
        del items
        self.indexed_item_processor = _IndexedItemProcessor(item_processor)
        self.action_name = action_name
        self.subject_name = subject_name
        self.keep_null_values = keep_null_values
        self.count_null_results = count_null_results
        self.count_failed_items = count_failed_items
        self.max_num_items = max_num_items
        self.logger = logger

        if thread_count == 0:
            thread_count = multiprocessing.cpu_count()
        if thread_count > len(self.items):
            thread_count = len(self.items)
        if max_num_items is not None and thread_count > max_num_items:
            thread_count = max_num_items
        if thread_count > 1:
            logger.debug(f"Processing items with {thread_count} subprocesses")
            original_sigint_handler = signal.signal(signal.SIGINT, signal.SIG_IGN)
            self.process_pool = multiprocessing.Pool(processes=thread_count)
            signal.signal(signal.SIGINT, original_sigint_handler)
        else:
            self.process_pool = None

        if action_name_past_form is None:
            if action_name.endswith("e"):
                action_name_past_form = action_name + "d"
            else:
                action_name_past_form = action_name + "ed"
        self.action_name_past_form = action_name_past_form
        self.processing_results: List[Optional[Any]] = list()

    def __drop_null_items(self):
        non_null_items = [item for item in self.items if item is not None]
        self.items_to_process_count = len(non_null_items)

        if not self.keep_null_values:
            self.items = non_null_items

    def __limit_number_of_items_to_process(self):
        if self.max_num_items is not None:
            self.items_to_process_count = min(self.items_to_process_count, self.max_num_items)
            if self.items_to_process_count < self.max_num_items:
                # limit doesn't get triggered
                self.max_num_items = None

    def __process_indexed_item_batch_in_parallel(self, indexed_item_batch: Iterable[Tuple[int, Optional[Any]]]) \
            -> Iterable[Tuple[int, Optional[Any]]]:
        try:
            indexed_results: Iterable[Tuple[int, Optional[Any]]] = self.process_pool.imap_unordered(
                self.indexed_item_processor,
                indexed_item_batch)
        except KeyboardInterrupt:
            self.logger.info("Terminating workers")
            self.process_pool.terminate()
            self.process_pool.join()
            raise
        else:
            self.process_pool.close()
            self.process_pool.join()
            return indexed_results

    def __process_indexed_item_batch(self, indexed_item_batch: Iterable[Tuple[int, Optional[Any]]]):
        if self.process_pool is None:
            indexed_results = [self.indexed_item_processor(indexed_item) for indexed_item in indexed_item_batch]
        else:
            indexed_results = self.__process_indexed_item_batch_in_parallel(indexed_item_batch)
        results: List[Optional[Any]] = [indexed_item[1] for indexed_item in sorted(indexed_results)]
        self.processing_results += results

    def __get_processing_results(self,
                                 keep_null_items: bool = False,
                                 keep_exceptions_as_nulls: bool = False) -> List[Any]:
        items = self.processing_results
        if not keep_null_items:
            items = [item for item in items if item is not None]
        if keep_exceptions_as_nulls:
            items = [None if isinstance(item, Exception) else item for item in items]
        else:
            items = [item for item in items if not isinstance(item, Exception)]
        return items

    def count_successfully_processed_items(self) -> int:
        return len(self.__get_processing_results(keep_null_items=self.count_null_results,
                                                 keep_exceptions_as_nulls=self.count_failed_items))

    def __process_items(self) -> None:
        indexed_items = list(enumerate(self.items))
        if self.max_num_items is None:
            self.__process_indexed_item_batch(indexed_items)
        else:
            while self.items_to_process_count > 0:
                processed_items_count = len(self.processing_results)
                current_item_batch = indexed_items[processed_items_count:
                                                   processed_items_count + self.items_to_process_count]
                if len(current_item_batch) == 0:
                    self.items_to_process_count = 0
                    break

                self.__process_indexed_item_batch(current_item_batch)

                self.items_to_process_count = self.max_num_items - self.count_successfully_processed_items()

    def __evaluate_results_textually(self) -> None:
        successfully_processed_items_count = self.count_successfully_processed_items()
        if successfully_processed_items_count > 0:
            self.logger.info(
                f"Successfully {self.action_name_past_form} {successfully_processed_items_count} {self.subject_name}"
                f"{utils.get_plural_s(successfully_processed_items_count)}")
        else:
            self.logger.info(f"No {self.subject_name}s were {self.action_name_past_form}")

    def process(self) -> Iterable[Optional[Any]]:
        self.__drop_null_items()
        if self.items_to_process_count == 0:
            self.logger.warning(f"No {self.subject_name}s available to {self.action_name}")
            return self.items

        self.__limit_number_of_items_to_process()
        self.logger.debug(f"Trying to {self.action_name} {self.items_to_process_count} {self.subject_name}"
                          f"{utils.get_plural_s(self.items_to_process_count)}")

        self.__process_items()
        self.__evaluate_results_textually()

        return self.__get_processing_results(keep_null_items=self.keep_null_values,
                                             keep_exceptions_as_nulls=self.keep_null_values)



#!/usr/bin/python3
import mmproteo.utils.filters
from mmproteo.utils import pride, formats
import pandas as pd
try:
    from utils.defaults import *
except ModuleNotFoundError:
    from .utils.defaults import *


def store_get_project_info():
    with open(f"resources/{DEFAULT_PROJECT}_info.txt", "w") as file:
        file.write(pride.get_project_info(project_name=DEFAULT_PROJECT, api_versions=[DEFAULT_TEST_API]))


def test_get_project_info():
    with open(f"resources/{DEFAULT_PROJECT}_info.txt", "r") as file:
        expected_info = file.read()

    received_info = pride.get_project_info(project_name=DEFAULT_PROJECT, api_versions=[DEFAULT_TEST_API])
    assert expected_info == received_info


def store_get_project_files():
    pride.get_project_files(project_name=DEFAULT_PROJECT, api_versions=[DEFAULT_TEST_API])\
        .to_parquet(f"resources/{DEFAULT_PROJECT}_files.parquet")


def test_get_project_files():
    expected_files = pd.read_parquet(f"resources/{DEFAULT_PROJECT}_files.parquet")
    received_files = pride.get_project_files(project_name=DEFAULT_PROJECT, api_versions=[DEFAULT_TEST_API])
    pd._testing.assert_frame_equal(expected_files, received_files, check_exact=True)


def store_get_raw_project_files():
    project_files = pride.get_project_files(project_name=DEFAULT_PROJECT, api_versions=[DEFAULT_TEST_API])
    project_files = mmproteo.utils.filters.filter_files_df(files_df=project_files, file_extensions={"raw"})
    project_files.to_parquet(f"resources/{DEFAULT_PROJECT}_files_raw.parquet")


def test_list_raw_project_files():
    expected_files = pd.read_parquet(f"resources/{DEFAULT_PROJECT}_files_raw.parquet")
    received_files = pride.get_project_files(project_name=DEFAULT_PROJECT, api_versions=[DEFAULT_TEST_API])
    received_files = mmproteo.utils.filters.filter_files_df(files_df=received_files, file_extensions={"raw"})
    pd._testing.assert_frame_equal(expected_files, received_files, check_exact=True)


def store_get_mgf_mzid_project_files():
    project_files = pride.get_project_files(project_name=DEFAULT_PROJECT, api_versions=[DEFAULT_TEST_API])
    project_files = mmproteo.utils.filters.filter_files_df(files_df=project_files, file_extensions={"mgf", "mzid"})
    project_files.to_parquet(f"resources/{DEFAULT_PROJECT}_files_mgf_mzid.parquet")


def test_list_mgf_mzid_files():
    expected_files = pd.read_parquet(f"resources/{DEFAULT_PROJECT}_files_mgf_mzid.parquet")
    received_files = pride.get_project_files(project_name=DEFAULT_PROJECT, api_versions=[DEFAULT_TEST_API])
    received_files = mmproteo.utils.filters.filter_files_df(files_df=received_files, file_extensions={"mgf", "mzid"})
    pd._testing.assert_frame_equal(expected_files, received_files, check_exact=True)


def store_get_gz_project_files():
    project_files = pride.get_project_files(project_name=DEFAULT_PROJECT, api_versions=[DEFAULT_TEST_API])
    project_files = mmproteo.utils.filters.filter_files_df(files_df=project_files, file_extensions={"gz"})
    project_files.to_parquet(f"resources/{DEFAULT_PROJECT}_files_gz.parquet")


def test_list_gz_files():
    expected_files = pd.read_parquet(f"resources/{DEFAULT_PROJECT}_files_gz.parquet")
    received_files = pride.get_project_files(project_name=DEFAULT_PROJECT, api_versions=[DEFAULT_TEST_API])
    received_files = mmproteo.utils.filters.filter_files_df(files_df=received_files, file_extensions={"gz"})
    pd._testing.assert_frame_equal(expected_files, received_files, check_exact=True)


if __name__ == '__main__':
    store_get_project_info()
    store_get_project_files()
    store_get_raw_project_files()
    store_get_mgf_mzid_project_files()
    store_get_gz_project_files()



import os
from typing import Dict, Set, Optional, List

from mmproteo.utils import utils, log
from mmproteo.utils.config import Config
from mmproteo.utils.filters import AbstractFilterConditionNode, filter_files_list
from mmproteo.utils.formats import read
from mmproteo.utils.processing import ItemProcessor

_FILE_EXTRACTION_CONFIG: Dict[str, Dict[str, str]] = {
    "gz": {
        "command": 'gunzip "%s"'
    },
    "zip": {
        "command": 'unzip "%s"'
    },
}


def get_extractable_file_extensions() -> Set[str]:
    return set(_FILE_EXTRACTION_CONFIG.keys())


def get_string_of_extractable_file_extensions(extension_quote: str = Config.default_option_quote,
                                              separator: str = Config.default_option_separator) -> str:
    return utils.concat_set_of_options(options=get_extractable_file_extensions(),
                                       option_quote=extension_quote,
                                       separator=separator)


def extract_file_if_possible(filename: Optional[str],
                             skip_existing: bool = Config.default_skip_existing,
                             logger: log.Logger = log.DEFAULT_LOGGER) -> Optional[str]:
    if filename is None:
        return None

    extracted_filename, file_ext = read.separate_extension(filename, get_extractable_file_extensions())
    if len(file_ext) == 0:
        logger.debug(f"Cannot extract file '{extracted_filename}', unknown extension")
        return None

    extraction_command = _FILE_EXTRACTION_CONFIG[file_ext]["command"] % filename

    if skip_existing and os.path.isfile(extracted_filename):
        logger.info(f'Skipping extraction, because "{extracted_filename}" already exists')
        return extracted_filename

    logger.info("Extracting file using '%s'" % extraction_command)
    return_code = os.system(extraction_command)
    if return_code == 0:
        logger.info(f"Extracted file '{extracted_filename}'")
    else:
        logger.warning(f'Failed extracting file "{filename}" (return code = {return_code})')

    return extracted_filename


class _ArchiveFileProcessor:
    def __init__(self,
                 skip_existing: bool = Config.default_skip_existing,
                 logger: log.Logger = log.DEFAULT_LOGGER):
        self.skip_existing = skip_existing
        self.logger = logger

    def __call__(self, filename: Optional[str]) -> Optional[str]:
        return extract_file_if_possible(filename=filename, skip_existing=self.skip_existing, logger=self.logger)


def extract_files(filenames: List[Optional[str]],
                  skip_existing: bool = Config.default_skip_existing,
                  max_num_files: Optional[int] = None,
                  count_failed_files: bool = Config.default_count_failed_files,
                  count_skipped_files: bool = Config.default_count_skipped_files,
                  thread_count: int = Config.default_thread_count,
                  column_filter: Optional[AbstractFilterConditionNode] = None,
                  keep_null_values: bool = Config.default_keep_null_values,
                  pre_filter_files: bool = Config.default_pre_filter_files,
                  logger: log.Logger = log.DEFAULT_LOGGER) -> List[Optional[str]]:
    if pre_filter_files:
        filenames = filter_files_list(filenames=filenames,
                                      file_extensions=get_extractable_file_extensions(),
                                      max_num_files=max_num_files,
                                      column_filter=column_filter,
                                      keep_null_values=keep_null_values,
                                      drop_duplicates=not keep_null_values,
                                      sort=not keep_null_values,
                                      logger=logger)

    file_processor = _ArchiveFileProcessor(skip_existing=skip_existing, logger=logger)

    return list(ItemProcessor(items=filenames,
                              item_processor=file_processor,
                              action_name="extract",
                              subject_name="archive file",
                              max_num_items=max_num_files,
                              count_failed_items=count_failed_files,
                              count_null_results=count_skipped_files,
                              thread_count=thread_count,
                              keep_null_values=keep_null_values,
                              logger=logger).process())



import os
from setuptools import setup, find_packages
import versioneer

with open(os.path.join('..', 'README.md'), "r") as file:
    read_me = file.read()

setup(
    name='mmproteo',
    version=versioneer.get_version(),
    description='Mirko meets Proteomics: A PRIDE downloader on steroids for deep learning-based DeNovo Sequencing',
    long_description=read_me,
    long_description_content_type="text/markdown",
    url='https://gitlab.com/dacs-hpi/pride-downloader',
    author='Mirko Krause',
    author_email='krause@codebase.one',
    license='GPLv3+',
    packages=find_packages(),
    entry_points={
        "console_scripts": ['mmproteo = mmproteo.mmproteo:main']
    },
    classifiers=[
        "Programming Language :: Python :: 3",
        "Environment :: Console",
        "License :: OSI Approved :: GNU General Public License v3 or later (GPLv3+)",
        "Operating System :: OS Independent",
    ],
    python_requires='>=3.6',
    install_requires=[
        "requests",
        "pandas",
        "pyteomics",
        "wget",
        "pyarrow",
        "lxml",
        "numpy"
    ],
    cmdclass=versioneer.get_cmdclass(),
)



from typing import Set

from mmproteo.utils import utils, commands, config, log
from .utils.defaults import *
from .utils.fixtures import *


def get_files_in_directory(directory_name: str) -> Set[str]:
    directory, directories, files = next(os.walk(directory_name))
    return set(files)


def _run_mz_pipeline_combined(conf: config.Config, logger: log.Logger):
    logger.fail_early = True
    conf.commands = ["list", "download", "extract", "mz2parquet"]
    commands.DISPATCHER.dispatch_commands(config=conf,
                                          catch_validation_warnings=False,
                                          catch_run_warnings=False,
                                          logger=logger)

    assert os.path.isfile(MZML_FILE), \
        "the mzml file should have been downloaded and extracted"
    assert os.path.isfile(MZID_FILE), \
        "the mzid file should have been downloaded and extracted"
    assert os.path.isfile(MZMLID_FILE), \
        "the mzml file and the mzid file should have been merged"

    assert get_files_in_directory(".") == {MZML_FILE, MZID_FILE, MZMLID_FILE}, \
        "there should only be the mzML, mzid, and parquet file"

    assert conf._project_files is not None, "the project files should have been cached"
    assert len(conf._project_files) == DEFAULT_PROJECT_FILE_COUNT, \
        f"there should be {DEFAULT_PROJECT_FILE_COUNT} cached project files"

    assert conf.default_downloaded_files_column in conf._processed_files, \
        "there should be references to downloaded files"
    assert len(conf._processed_files[conf.default_downloaded_files_column].dropna()) == conf.max_num_files, \
        f"there should be exactly {conf.max_num_files} referenced downloaded files"

    assert conf.default_extracted_files_column in conf._processed_files, \
        "there should be references to extracted files"
    assert len(conf._processed_files[conf.default_extracted_files_column].dropna()) == conf.max_num_files, \
        f"there should be exactly {conf.max_num_files} referenced extracted files"

    assert conf.default_mzmlid_parquet_files_column in conf._processed_files, \
        f"there should be references to converted '*{conf.default_mzmlid_parquet_file_postfix}' files"
    assert len(conf._processed_files[conf.default_mzmlid_parquet_files_column].dropna()) == conf.max_num_files / 2, \
        f"there should be exactly {conf.max_num_files/2} referenced converted " \
        f"'*{conf.default_mzmlid_parquet_file_postfix}' file"

    conf.clear_cache()


def _run_mz_pipeline_in_single_steps_with_full_cache(conf: config.Config, logger: log.Logger):
    logger.fail_early = False
    conf.commands = ["download"]
    commands.DISPATCHER.dispatch_commands(config=conf,
                                          catch_validation_warnings=False,
                                          catch_run_warnings=False,
                                          logger=logger)

    downloaded_files = utils.merge_column_values(conf._processed_files, [conf.default_downloaded_files_column])
    assert len(downloaded_files) == 0, "there should be no downloaded files in the second iteration"

    conf.clear_cache()
    conf.commands = ["extract"]
    commands.DISPATCHER.dispatch_commands(config=conf,
                                          catch_validation_warnings=False,
                                          catch_run_warnings=False,
                                          logger=logger)

    extracted_files = utils.merge_column_values(conf._processed_files, [conf.default_extracted_files_column])
    assert len(extracted_files) == 0, "there should be no extracted files in the second iteration"

    conf.clear_cache()
    conf.commands = ["mz2parquet"]
    commands.DISPATCHER.dispatch_commands(config=conf,
                                          catch_validation_warnings=False,
                                          catch_run_warnings=False,
                                          logger=logger)

    downloaded_and_extracted_files = utils.merge_column_values(conf._processed_files,
                                                               [conf.default_mzmlid_parquet_files_column])

    assert len(downloaded_and_extracted_files) == 0, "there should be no converted files in the second iteration"

    assert get_files_in_directory(".") == {MZML_FILE, MZID_FILE, MZMLID_FILE}, \
        "there should only be the mzML, mzid, and parquet file"


def test_mz_pipeline_single_threaded(run_in_temp_directory):
    conf = config.Config()
    conf.pride_project = DEFAULT_PROJECT
    conf.valid_file_extensions = ["mzid", "mzml"]
    conf.max_num_files = 2
    conf.shown_columns = ["fileName", "fileSizeBytes"]
    conf.thread_count = 1

    logger = log.TestLogger(terminate_process=False,
                            verbose=True)
    conf.check(logger)

    _run_mz_pipeline_combined(conf=conf, logger=logger)
    _run_mz_pipeline_in_single_steps_with_full_cache(conf=conf, logger=logger)


def test_mz_pipeline_in_parallel(run_in_temp_directory):
    conf = config.Config()
    conf.pride_project = DEFAULT_PROJECT
    conf.valid_file_extensions = ["mzid", "mzml"]
    conf.max_num_files = 2
    conf.shown_columns = ["fileName", "fileSizeBytes"]
    conf.thread_count = 2

    logger = log.TestLogger(terminate_process=False,
                            verbose=True)
    conf.check(logger)

    _run_mz_pipeline_combined(conf=conf, logger=logger)
    _run_mz_pipeline_in_single_steps_with_full_cache(conf=conf, logger=logger)


def test_mz_pipeline_in_single_steps(run_in_temp_directory):
    conf = config.Config()
    conf.pride_project = DEFAULT_PROJECT
    conf.max_num_files = 1

    logger = log.TestLogger(fail_early=True,
                            terminate_process=False,
                            verbose=True)
    conf.check(logger=logger)

    conf.commands = ["download"]
    conf.valid_file_extensions = ["mzml"]
    commands.DISPATCHER.dispatch_commands(config=conf,
                                          catch_validation_warnings=False,
                                          catch_run_warnings=False,
                                          logger=logger)
    conf.clear_cache()

    assert os.path.isfile(MZML_GZ_FILE), \
        "the mzML.gz file should have been downloaded"

    conf.commands = ["download"]
    conf.valid_file_extensions = ["mzid"]
    commands.DISPATCHER.dispatch_commands(config=conf,
                                          catch_validation_warnings=False,
                                          catch_run_warnings=False,
                                          logger=logger)
    conf.clear_cache()

    assert os.path.isfile(MZID_GZ_FILE), \
        "the mzid.gz file should have been downloaded"

    conf.commands = ["extract"]
    conf.valid_file_extensions = ["mzml"]
    commands.DISPATCHER.dispatch_commands(config=conf,
                                          catch_validation_warnings=False,
                                          catch_run_warnings=False,
                                          logger=logger)
    conf.clear_cache()

    assert os.path.isfile(MZML_FILE), \
        "the mzML file should have been extracted"

    assert not os.path.isfile(MZML_GZ_FILE), \
        "the mzML.gz file should have been deleted by gzip"

    conf.commands = ["extract"]
    conf.valid_file_extensions = ["mzid"]
    commands.DISPATCHER.dispatch_commands(config=conf,
                                          catch_validation_warnings=False,
                                          catch_run_warnings=False,
                                          logger=logger)
    conf.clear_cache()

    assert os.path.isfile(MZID_FILE), \
        "the mzid file should have been extracted"

    assert not os.path.isfile(MZID_GZ_FILE), \
        "the mzid.gz file should have been deleted by gzip"

    conf.commands = ["mz2parquet"]
    conf.valid_file_extensions = []
    commands.DISPATCHER.dispatch_commands(config=conf,
                                          catch_validation_warnings=False,
                                          catch_run_warnings=False,
                                          logger=logger)
    conf.clear_cache()

    assert os.path.isfile(MZMLID_FILE), \
        "the mzml file and the mzid file should have been merged"

    assert get_files_in_directory(".") == {MZML_FILE, MZID_FILE, MZMLID_FILE}, \
        "there should only be the mzML, mzid, and parquet file"


def _run_raw_mgf_pipeline_combined(conf: Config, logger: log.Logger):
    logger.fail_early = True
    conf.commands = ["download", "convertraw", "mgf2parquet"]
    commands.DISPATCHER.dispatch_commands(config=conf,
                                          catch_validation_warnings=False,
                                          catch_run_warnings=False,
                                          logger=logger)

    assert os.path.isfile(RAW_FILE), \
        "the raw file should have been downloaded"
    assert os.path.isfile(MGF_FILE), \
        "the raw file should have been converted to mgf format"
    assert os.path.isfile(MGF_PARQUET_FILE), \
        "the mgf file should have been converted to parquet format"

    assert conf.default_downloaded_files_column in conf._processed_files, \
        "there should be references to downloaded files"
    assert len(conf._processed_files[conf.default_downloaded_files_column].dropna()) == conf.max_num_files, \
        f"there should be exactly {conf.max_num_files} referenced downloaded file"

    assert conf.default_converted_raw_files_column in conf._processed_files, \
        "there should be references to converted raw files"
    assert len(conf._processed_files[conf.default_converted_raw_files_column].dropna()) == conf.max_num_files, \
        f"there should be exactly {conf.max_num_files} referenced converted raw file"

    assert conf.default_mgf_parquet_files_column in conf._processed_files, \
        "there should be references to converted mgf files in parquet format"
    assert len(conf._processed_files[conf.default_mgf_parquet_files_column].dropna()) == conf.max_num_files, \
        f"there should be exactly {conf.max_num_files} referenced converted mgf files in parquet format"

    if conf.max_num_files == 1:
        assert get_files_in_directory(".") == {RAW_FILE, MGF_FILE, MGF_PARQUET_FILE}, \
            "there should only be the mzML, mzid, and parquet file"


def test_raw_mgf_pipeline_single_threaded(run_in_temp_directory):
    conf = config.Config()
    conf.pride_project = DEFAULT_PROJECT
    conf.valid_file_extensions = ["raw"]
    conf.max_num_files = 1
    conf.thread_count = 1

    logger = log.TestLogger(terminate_process=False,
                            verbose=True)
    conf.check(logger)

    _run_raw_mgf_pipeline_combined(conf=conf, logger=logger)


def test_raw_mgf_pipeline_parallelized(run_in_temp_directory):
    conf = config.Config()
    conf.pride_project = DEFAULT_PROJECT
    conf.valid_file_extensions = ["raw"]
    conf.max_num_files = 2
    conf.thread_count = 2

    logger = log.TestLogger(terminate_process=False,
                            verbose=True)
    conf.check(logger)

    _run_raw_mgf_pipeline_combined(conf=conf, logger=logger)



import os
from typing import List, Optional, Tuple, Dict, Any

import pandas as pd
from pyteomics import mzid, mzml

from mmproteo.utils import log, utils
from mmproteo.utils.config import Config
from mmproteo.utils.filters import AbstractFilterConditionNode, filter_files_list
from mmproteo.utils.formats import read
from mmproteo.utils.processing import ItemProcessor


def merge_mzml_and_mzid_dfs(mzml_df: pd.DataFrame,
                            mzid_df: pd.DataFrame,
                            mzml_key_columns: List[str] = None,
                            mzid_key_columns: List[str] = None,
                            logger: log.Logger = log.DEFAULT_LOGGER) -> pd.DataFrame:
    if mzml_key_columns is None:
        mzml_key_columns = Config.default_mzml_key_columns
    if mzid_key_columns is None:
        mzid_key_columns = Config.default_mzid_key_columns

    input_length = min(len(mzml_df), len(mzid_df))

    logger.debug("Started merging MzML and MzID dataframes")
    merged_df = mzml_df.merge(right=mzid_df,
                              how='inner',
                              left_on=mzml_key_columns,
                              right_on=mzid_key_columns)
    output_length = len(merged_df)
    unmatched_rows = input_length - output_length
    if unmatched_rows < 0:
        if logger.is_verbose():
            unique_mzml_length = len(mzml_df.drop_duplicates(inplace=False, subset=mzml_key_columns))
            unique_mzid_length = len(mzid_df.drop_duplicates(inplace=False, subset=mzid_key_columns))
            duplicate_mzml_rows = len(mzml_df) - unique_mzml_length
            duplicate_mzid_rows = len(mzid_df) - unique_mzid_length
            logger.debug("Duplicate MzML rows = %d" % duplicate_mzml_rows)
            logger.debug("Duplicate MzID rows = %d" % duplicate_mzid_rows)
        logger.warning("The key columns were no real key columns, because duplicates were found.")

    logger.debug("Finished merging MzML and MzID dataframes and matched %d x %d -> %d rows" %
                 (len(mzml_df), len(mzid_df), output_length))
    return merged_df


def merge_mzml_and_mzid_files(mzml_filename: str,
                              mzid_filename: str,
                              mzml_key_columns: Optional[List[str]] = None,
                              mzid_key_columns: Optional[List[str]] = None,
                              logger: log.Logger = log.DEFAULT_LOGGER) -> pd.DataFrame:
    if mzml_key_columns is None:
        mzml_key_columns = Config.default_mzml_key_columns
    if mzid_key_columns is None:
        mzid_key_columns = Config.default_mzid_key_columns

    logger.debug("Started Merge: '%s' + '%s' -> dataframe" % (mzml_filename, mzid_filename))
    mzml_df = read.read(filename=mzml_filename, logger=logger)
    mzid_df = read.read(filename=mzid_filename, logger=logger)

    merged_df = merge_mzml_and_mzid_dfs(mzml_df=mzml_df,
                                        mzid_df=mzid_df,
                                        mzml_key_columns=mzml_key_columns,
                                        mzid_key_columns=mzid_key_columns,
                                        logger=logger)
    logger.debug("Finished Merge: '%s' + '%s' -> dataframe" % (mzml_filename, mzid_filename))
    return merged_df


def _process_mzml_and_mzid_to_parquet_merge_job(mzml_filename: str,
                                                mzid_filename: str,
                                                target_filename: str,
                                                current_merge_job_index: int,
                                                merge_job_count: int,
                                                skip_existing: bool = Config.default_skip_existing,
                                                mzml_key_columns: Optional[List[str]] = None,
                                                mzid_key_columns: Optional[List[str]] = None,
                                                logger: log.Logger = log.DEFAULT_LOGGER) -> Optional[str]:
    if skip_existing and os.path.exists(target_filename):
        logger.info("Skipping Merge %d/%d: '%s' + '%s' -> '%s' already exists" %
                    (current_merge_job_index + 1, merge_job_count, mzml_filename, mzid_filename, target_filename))
        return None

    logger.info("Started Merge %d/%d: '%s' + '%s' -> '%s'" %
                (current_merge_job_index + 1, merge_job_count, mzml_filename, mzid_filename, target_filename))
    merged_df = merge_mzml_and_mzid_files(mzml_filename=mzml_filename,
                                          mzid_filename=mzid_filename,
                                          mzml_key_columns=mzml_key_columns,
                                          mzid_key_columns=mzid_key_columns,
                                          logger=logger)
    merged_df.to_parquet(path=target_filename)
    logger.info("Finished Merge %d/%d: '%s' + '%s' -> '%s'" %
                (current_merge_job_index + 1, merge_job_count, mzml_filename, mzid_filename, target_filename))

    return target_filename


class _Mz2ParquetMergeJobProcessor:
    def __init__(self,
                 merge_job_count: int,
                 skip_existing: bool = Config.default_skip_existing,
                 mzml_key_columns: Optional[List[str]] = None,
                 mzid_key_columns: Optional[List[str]] = None,
                 logger: log.Logger = log.DEFAULT_LOGGER):
        self.merge_job_count = merge_job_count
        self.skip_existing = skip_existing
        self.mzml_key_columns = mzml_key_columns
        self.mzid_key_columns = mzid_key_columns
        self.logger = logger

    def __call__(self, merge_job: Tuple[int, Tuple[str, str, str]]) -> Optional[str]:
        index, (mzml_filename, mzid_filename, target_filename) = merge_job
        return _process_mzml_and_mzid_to_parquet_merge_job(mzml_filename=mzml_filename,
                                                           mzid_filename=mzid_filename,
                                                           target_filename=target_filename,
                                                           current_merge_job_index=index,
                                                           merge_job_count=self.merge_job_count,
                                                           skip_existing=self.skip_existing,
                                                           mzml_key_columns=self.mzml_key_columns,
                                                           mzid_key_columns=self.mzid_key_columns,
                                                           logger=self.logger)


def _create_merge_jobs(filenames_and_extensions: List[Tuple[str, Tuple[str, str]]],
                       prefix_length_tolerance: int,
                       target_filename_postfix: str,
                       skip_existing: bool = Config.default_skip_existing,
                       logger: log.Logger = log.DEFAULT_LOGGER) -> List[Tuple[str, str, str]]:
    filenames_and_extensions = sorted(filenames_and_extensions)
    merge_jobs = []

    last_filename, (last_filename_prefix, last_extension) = filenames_and_extensions[0]
    for filename, (filename_prefix, extension) in filenames_and_extensions[1:]:
        if last_filename is not None and extension != last_extension:
            common_filename_prefix_length = len(os.path.commonprefix([filename_prefix, last_filename_prefix]))
            required_filename_prefix_length = min(len(filename_prefix),
                                                  len(last_filename_prefix)) - prefix_length_tolerance
            if common_filename_prefix_length >= required_filename_prefix_length:
                # found a possible merging pair
                if extension == "mzml":
                    mzml_filename = filename
                    mzid_filename = last_filename
                else:
                    mzml_filename = last_filename
                    mzid_filename = filename

                target_filename = filename[:common_filename_prefix_length] + target_filename_postfix
                if skip_existing and os.path.exists(target_filename):
                    logger.info(f"Skipping Merge: '{mzml_filename}' + '{mzid_filename}' "
                                f"-> '{target_filename}' already exists")
                else:
                    merge_jobs.append((mzml_filename, mzid_filename, target_filename))

                filename = None  # skip next iteration to prevent merging the same file with multiple others
        last_filename = filename
        last_filename_prefix = filename_prefix
        last_extension = extension

    return merge_jobs


def merge_mzml_and_mzid_files_to_parquet(filenames: List[Optional[str]],
                                         skip_existing: bool = Config.default_skip_existing,
                                         max_num_files: Optional[int] = None,
                                         count_failed_files: bool = Config.default_count_failed_files,
                                         count_skipped_files: bool = Config.default_count_skipped_files,
                                         column_filter: Optional[AbstractFilterConditionNode] = None,
                                         mzml_key_columns: Optional[List[str]] = None,
                                         mzid_key_columns: Optional[List[str]] = None,
                                         prefix_length_tolerance: int = 0,
                                         target_filename_postfix: str = Config.default_mzmlid_parquet_file_postfix,
                                         thread_count: int = Config.default_thread_count,
                                         logger: log.Logger = log.DEFAULT_LOGGER) -> List[str]:
    filenames = filter_files_list(filenames=filenames,
                                  column_filter=column_filter,
                                  sort=False,
                                  logger=logger)

    filenames_and_extensions: List[Tuple[str, Tuple[str, str]]] \
        = [(filename, read.separate_extension(filename=filename, extensions={"mzml", "mzid"}))
           for filename in filenames if filename is not None]
    filenames_and_extensions = [(filename, (file, ext)) for filename, (file, ext) in filenames_and_extensions if
                                len(ext) > 0]
    if len(filenames_and_extensions) < 2:
        logger.warning("No MzML and MzID files available for merging")
        return []

    merge_jobs = _create_merge_jobs(filenames_and_extensions=filenames_and_extensions,
                                    prefix_length_tolerance=prefix_length_tolerance,
                                    target_filename_postfix=target_filename_postfix,
                                    skip_existing=skip_existing,
                                    logger=logger)

    item_processor = _Mz2ParquetMergeJobProcessor(merge_job_count=len(merge_jobs),
                                                  skip_existing=skip_existing,
                                                  mzml_key_columns=mzml_key_columns,
                                                  mzid_key_columns=mzid_key_columns,
                                                  logger=logger)

    parquet_files = list(ItemProcessor(items=enumerate(merge_jobs),
                                       item_processor=item_processor,
                                       action_name="mzmlid-merge",
                                       subject_name="file pair",
                                       max_num_items=max_num_files,
                                       count_failed_items=count_failed_files,
                                       count_null_results=count_skipped_files,
                                       keep_null_values=False,
                                       thread_count=thread_count,
                                       logger=logger).process())
    return parquet_files


def _prepare_mzid_entry(entry: Dict[str, Any]) -> Dict[str, Any]:
    entry = entry.copy()

    try:
        items = utils.list_of_dicts_to_dict(entry["SpectrumIdentificationItem"], "rank")
        if items is not None:
            entry["SpectrumIdentificationItem"] = items
    except:
        pass

    entry = utils.flatten_dict(entry)
    return entry


def read_mzid(filename: str, logger: log.Logger = log.DEFAULT_LOGGER) -> pd.DataFrame:
    entries = read.iter_entries(mzid.read(filename), logger=logger)
    extracted_entries = [_prepare_mzid_entry(entry) for entry in entries]
    return pd.DataFrame(data=extracted_entries)


def read_mzml(filename: str, logger: log.Logger = log.DEFAULT_LOGGER) -> pd.DataFrame:
    entries = read.iter_entries(mzml.read(filename), logger=logger)
    extracted_entries = [utils.flatten_dict(entry) for entry in entries]
    return pd.DataFrame(data=extracted_entries)



import argparse
from operator import attrgetter
from typing import Any, List, Optional, Set, Tuple, Union

import pandas as pd

import mmproteo.utils.filters
from mmproteo._version import get_versions
from mmproteo.utils import log, utils

__version__ = get_versions()['version']

del get_versions


class _MultiLineArgumentDefaultsHelpFormatter(argparse.ArgumentDefaultsHelpFormatter):

    def add_arguments(self, actions):
        actions = sorted(actions, key=attrgetter('option_strings'))
        super(_MultiLineArgumentDefaultsHelpFormatter, self).add_arguments(actions)

    def _format_usage(self, usage, actions, groups, prefix):
        actions = sorted(actions, key=attrgetter('option_strings'))
        return super(_MultiLineArgumentDefaultsHelpFormatter, self)._format_usage(usage=usage,
                                                                                  actions=actions,
                                                                                  groups=groups,
                                                                                  prefix=prefix)

    def _split_lines(self, text: str, width: int) -> List[str]:
        lines = text.splitlines()
        wrapped_lines = []
        import textwrap
        for line in lines:
            command_description_parts = line.split(" : ")
            if len(command_description_parts) == 2 and command_description_parts[0].rstrip(" ").isalnum():
                description_width = width - len(command_description_parts[0]) - len(" : ")
                wrapped_description = textwrap.wrap(command_description_parts[1], description_width)
                wrapped_lines.append(command_description_parts[0] + " : " + wrapped_description[0])
                wrapped_lines += [" " * (width - description_width) + description_line
                                  for description_line in wrapped_description[1:]]
            else:
                wrapped_lines += textwrap.wrap(line, width)
        return wrapped_lines


class Config:
    default_application_name = "mmproteo"

    _pride_project_parameter_str: str = "--pride-project"

    default_storage_dir: str = "."
    default_log_file: str = default_application_name + ".log"

    default_file_name_column: str = "fileName"
    default_download_link_column: str = 'downloadLink'
    default_downloaded_files_column: str = 'downloaded_files'
    default_extracted_files_column: str = 'extracted_files'
    default_converted_raw_files_column: str = 'converted_raw_files'
    default_mzmlid_parquet_files_column: str = 'converted_mzmlid_parquet_files'
    default_mgf_parquet_files_column: str = 'converted_mgf_parquet_files'

    special_column_names: Set[str] = {
        default_file_name_column,
        default_download_link_column,
        default_downloaded_files_column,
        default_extracted_files_column,
        default_converted_raw_files_column,
        default_mzmlid_parquet_files_column,
        default_mgf_parquet_files_column
    }

    default_thermo_docker_container_name: str = "thermorawfileparser"
    default_thermo_docker_image: str = "quay.io/biocontainers/thermorawfileparser:1.3.2--h1341992_1"
    default_thermo_start_container_command_template: str = \
        "docker run --rm -w /data -v {abs_storage_dir}:/data --name {container_name} -d {image_name} tail -f /dev/null"
    default_thermo_output_format: str = "mgf"
    default_thermo_exec_command: str = "docker exec -i {container_name} ThermoRawFileParser -f {format} " \
                                       "-i /data/{input} -o /data"
    default_thermo_keep_container_running: bool = False
    default_option_quote: str = '"'
    default_option_separator: str = ", "
    default_skip_existing: bool = True
    default_filter_sort: bool = True
    default_filter_drop_duplicates: bool = True
    default_filter_separator_regex: str = "[=!]="
    default_filter_or_separator: str = " or "
    default_count_failed_files: bool = False
    default_count_null_results: bool = False
    default_count_skipped_files: bool = True
    default_keep_null_values: bool = False
    default_pre_filter_files: bool = True
    default_mzml_key_columns: List[str] = ['mzml_filename', 'id']
    default_mzid_key_columns: List[str] = ['name', 'spectrumID']
    default_mzmlid_parquet_file_postfix: str = "_mzmlid.parquet"
    default_thread_count: int = 1

    def __init__(self, logger: log.Logger = log.DEFAULT_LOGGER):
        self._logger = logger

        from mmproteo.utils import formats

        self.pride_project: Optional[str] = None
        self.max_num_files: int = 0
        self.count_failed_files: bool = self.default_count_failed_files
        self.count_skipped_files: bool = self.default_count_skipped_files
        self.storage_dir: str = self.default_storage_dir
        self.log_file: str = self.default_log_file
        self.log_to_stdout: bool = False
        self.valid_file_extensions: List[str] = []
        self.skip_existing: bool = self.default_skip_existing
        self.verbose: bool = False
        self.dummy_logger: bool = False
        self.shown_columns: List[str] = []
        self.commands: Optional[List[str]] = None
        self.pride_versions: List[str] = []
        self.column_filter: Optional[mmproteo.utils.filters.AbstractFilterConditionNode] = None
        self.fail_early: bool = True
        self.terminate_process: bool = False
        self.thermo_output_format: str = self.default_thermo_output_format
        self.thermo_keep_container_running: bool = self.default_thermo_keep_container_running
        self.thread_count = self.default_thread_count

        # cache
        self._processed_files: Optional[pd.DataFrame] = None
        self._project_files: Optional[pd.DataFrame] = None

    def set_logger(self, logger: log.Logger) -> None:
        self._logger = logger

    def get_processed_files(self, *columns: str) -> List[str]:
        return utils.merge_column_values(df=self._processed_files, columns=columns)

    def get_project_files(self) -> Optional[pd.DataFrame]:
        if self._project_files is None:
            from mmproteo.utils import pride
            self._project_files = pride.get_project_files(project_name=self.pride_project,
                                                          api_versions=self.pride_versions,
                                                          logger=self._logger)
        return self._project_files

    @staticmethod
    def get_string_of_special_column_names(extension_quote: str = default_option_quote,
                                           separator: str = default_option_separator) -> str:
        return utils.concat_set_of_options(options=Config.special_column_names,
                                           option_quote=extension_quote,
                                           separator=separator)

    def clear_cache(self):
        self._processed_files = None
        self._project_files = None

    @staticmethod
    def _get_negation_argument_prefix(condition: bool, negation_str: str = 'no-') -> str:
        if condition:
            return ""
        return negation_str

    def cache_processed_files(self,
                              data_list: Optional[List[Any]] = None,
                              column_names: Optional[Union[str, List[str]]] = None,
                              data_df: pd.DataFrame = None) -> Optional[pd.DataFrame]:
        assert (data_list is not None and column_names is not None and data_df is None) \
               or (data_list is None and column_names is None), "data_list and column_names must always be given " \
                                                                "together, but never at the same time as data_df"

        if data_df is None:
            if len(data_list) == 0:
                return None

            if type(column_names) == str:
                column_names = [column_names]

            data_df = pd.DataFrame(data=data_list, columns=column_names)

        if self._processed_files is None:
            self._processed_files = data_df
        else:
            self._processed_files = self._processed_files.append(data_df, ignore_index=True)
        return data_df

    def parse_arguments(self) -> None:
        from mmproteo.utils import commands, pride, utils
        from mmproteo.utils.formats.raw import get_thermo_raw_file_parser_output_formats
        parser = argparse.ArgumentParser(formatter_class=_MultiLineArgumentDefaultsHelpFormatter, add_help=False)

        parser.add_argument("command",
                            nargs='+',
                            choices=commands.DISPATCHER.get_command_names(),
                            metavar="COMMAND",
                            help="the list of actions to be performed on the repository. " +
                                 "Every action can only occur once. " +
                                 "Duplicates are dropped after the first occurrence.\n \n" +
                                 commands.DISPATCHER.get_command_descriptions_str())
        parser.add_argument(self._pride_project_parameter_str, "-p",
                            metavar="PROJECT",
                            help="the name of the PRIDE project, e.g. 'PXD010000' " +
                                 "from 'https://www.ebi.ac.uk/pride/ws/archive/peptide/list/project/PXD010000'. " +
                                 "For some commands, this parameter is required.")
        parser.add_argument("--max-num-files", "-n",
                            metavar="N",
                            default=self.max_num_files,
                            type=int,
                            help="the maximum number of files to be downloaded. Set it to '0' to download all files.")
        parser.add_argument(f"--{self._get_negation_argument_prefix(not self.count_failed_files)}count-failed-files",
                            action="store_" + str(self.count_failed_files).lower(),
                            dest='count_failed_files',
                            default=self.count_failed_files,
                            help=("Count failed files and do not just ignore them. " if not self.count_failed_files else
                                  "Do not count failed files and just ignore them. ") +
                                 "This is relevant for the max-num-files parameter.")
        parser.add_argument(f"--{self._get_negation_argument_prefix(not self.count_skipped_files)}count-skipped-files",
                            action="store_" + str(self.count_skipped_files).lower(),
                            dest='count_skipped_files',
                            default=self.count_skipped_files,
                            help=("Count skipped files and do not just ignore them. " if not self.count_skipped_files
                                  else "Do not count skipped files and just ignore them. ") +
                                 "This is relevant for the max-num-files parameter.")
        parser.add_argument("--storage-dir", "-d",
                            metavar="DIR",
                            default=self.storage_dir,
                            help="the name of the directory, in which the downloaded files and the log file will be "
                                 "stored.")
        parser.add_argument("--log-file", "-l",
                            metavar="FILE",
                            default=self.log_file,
                            help="the name of the log file, relative to the download directory. "
                                 "Set it to an empty string (\"\") to disable file logging.")
        parser.add_argument("--log-to-stdout",
                            action="store_" + str(not self.log_to_stdout).lower(),
                            help="Log to stdout instead of stderr.")
        parser.add_argument("--shown-columns", "-c",
                            metavar="COLUMNS",
                            default="",
                            type=lambda s: [col for col in s.split(",") if len(col) > 0],
                            help="a list of comma-separated column names. Some commands show their results as tables, "
                                 "so their output columns will be limited to those in this list. An empty list "
                                 "deactivates filtering. Capitalization matters.")
        parser.add_argument("--file-extensions", "-e",
                            metavar="EXT",
                            default="",
                            type=lambda s: {ext.lower() for ext in s.split(',') if len(ext) > 0},
                            help="a list of comma-separated allowed file extensions to filter files for. "
                                 "Archive extensions will be automatically appended. " +
                                 "An empty list deactivates filtering. "
                                 "Capitalization does not matter.")
        parser.add_argument("--no-skip-existing",
                            action="store_" + str(self.skip_existing).lower(),
                            help="Do not skip existing files.")
        # store_true turns "verbose" into a flag:
        # The existence of "verbose" equals True, the lack of existence equals False
        parser.add_argument("--verbose", "-v",
                            action="store_" + str(not self.verbose).lower(),
                            help="Increase output verbosity to debug level.")
        parser.add_argument("--no-fail-early",
                            action="store_" + str(self.fail_early).lower(),
                            help="Do not fail commands already on failed assertions. The code will run until "
                                 "a real exception is encountered or it even succeeds.")
        parser.add_argument('--help', "-h",
                            action='help',
                            default=argparse.SUPPRESS,
                            help='Show this help message and exit.')
        parser.add_argument('--version',
                            action='version',
                            version='%(prog)s ' + __version__,
                            help="Show the version of this software.")
        parser.add_argument('--pride-version', '-i',
                            choices=pride.get_pride_api_versions(),
                            action="append",
                            default=self.pride_versions,
                            help="an API version for the PRIDE interactions. Only the specified versions will be used. "
                                 "This parameter can be given multiple times to allow multiple different "
                                 "API versions, one version per parameter appearance. "
                                 "The order of occurring API versions will be considered until the first API request "
                                 "fulfills its job. Every version should appear at most once. "
                                 "Duplicates are dropped after the first occurrence. "
                                 "An empty list (default) uses all api versions in the following order: [%s]" %
                                 pride.get_string_of_pride_api_versions())
        parser.add_argument("--dummy-logger",
                            action="store_" + str(not self.dummy_logger).lower(),
                            help="Use a simpler log format and log to stdout.")
        parser.add_argument("--thermo-output-format",
                            default=self.thermo_output_format,
                            choices=get_thermo_raw_file_parser_output_formats(),
                            help="the output format into which the raw file will be converted. This parameter only "
                                 f"applies to the {commands.ConvertRawCommand().get_command()} command.")
        parser.add_argument("--thermo-keep-running",
                            action="store_" + str(not self.thermo_keep_container_running).lower(),
                            help="Keep the ThermoRawFileParser Docker container running after conversion. This can "
                                 "speed up batch processing and ease debugging.")
        parser.add_argument('--filter', '-f',
                            metavar=f"COLUMN{self.default_filter_separator_regex}REGEX",
                            action="append",
                            type=mmproteo.utils.filters.create_or_filter_from_str,
                            default=[],
                            help="a filter condition for file filtering. The condition must be of the form "
                                 f"'columnName{self.default_filter_separator_regex}valueRegex'. Therefore, the "
                                 "comparison operator can either be '==' or '!='. "
                                 "The column name must not contain these character patterns. "
                                 "The value will be interpreted using Python's rules for regular expressions (from the "
                                 "Python 're' package). This parameter can be given multiple times to enforce multiple "
                                 "filters simultaneously, meaning the filters will be logically connected using a "
                                 "boolean 'and'. Boolean 'or' operations can be specified as "
                                 f"'{self.default_filter_or_separator}' within any filter parameter, "
                                 "for example like this (representing (a==1 or b==2) and (c==3 or (not d==4))): "
                                 f"'{self.default_application_name} -f \"a==1{self.default_filter_or_separator}b==2\" "
                                 f"-f \"c==3{self.default_filter_or_separator}d!=4\" list'. "
                                 "A condition can be negated using '!=' instead of '=='. For the filtering process, "
                                 "the filter columns will be converted to strings. Non-existent columns will be "
                                 "ignored. Capitalization matters. All these rules add up to a "
                                 "conjunctive normal form. As some commands can be pipelined to use previous results, "
                                 "there are also the following special column names available: " +
                                 f"[{self.get_string_of_special_column_names()}]. An empty list disables this filter.")
        parser.add_argument("--thread-count",
                            metavar="THREADS",
                            default=self.thread_count,
                            type=int,
                            help="the number of threads to use for parallel processing. Set it to '0' to use as many "
                                 "threads as there are CPU cores. Setting the number of threads to '1' disables "
                                 "parallel processing.")

        args = parser.parse_args()

        self.pride_project = args.pride_project
        self.max_num_files = args.max_num_files
        self.count_failed_files = args.count_failed_files
        self.count_skipped_files = args.count_skipped_files
        self.storage_dir = args.storage_dir
        self.log_file = args.log_file
        self.log_to_stdout = args.log_to_stdout
        self.valid_file_extensions = args.file_extensions
        self.skip_existing = (not args.no_skip_existing)
        self.verbose = args.verbose
        self.dummy_logger = args.dummy_logger
        self.fail_early = (not args.no_fail_early)
        self.shown_columns = args.shown_columns
        self.pride_versions = utils.deduplicate_list(args.pride_version)
        self.column_filter = mmproteo.utils.filters.NoneFilterConditionNode(
            condition=mmproteo.utils.filters.AndFilterConditionNode(conditions=args.filter),
            none_value=True,
        )
        self.thermo_output_format = args.thermo_output_format
        self.thermo_keep_container_running = args.thermo_keep_running
        self.thread_count = args.thread_count

        self.commands = utils.deduplicate_list(args.command)

    def require_pride_project(self, logger: Optional[log.Logger] = None) -> None:
        if logger is None:
            logger = self._logger
        logger.assert_true(self.pride_project is not None, Config._pride_project_parameter_str + " is missing")
        logger.assert_true(len(self.pride_project) > 0,
                           Config._pride_project_parameter_str + " must not be empty")

    def validate_arguments(self, logger: Optional[log.Logger] = None) -> None:
        if logger is None:
            logger = self._logger
        logger.assert_true(self.storage_dir is None or len(self.storage_dir) > 0, "storage-dir must not be empty")
        logger.assert_true(self.max_num_files >= 0, "max-num-files must be >= 0; use 0 to process all files")
        logger.assert_true(self.thread_count >= 0,
                           "thread-count must be >= 0; use 0 to automatically set the number "
                           "of threads")
        if not self.skip_existing and self.count_skipped_files:
            logger.warning("skip-existing is not set although count-skipped-files is set")

    def check(self, logger: Optional[log.Logger] = None) -> None:
        if logger is None:
            logger = self._logger
        from mmproteo.utils import utils
        utils.ensure_dir_exists(self.storage_dir, logger=logger)

    @staticmethod
    def __sort_if_set(obj: Optional[Union[Set, Any]]) -> Optional[Union[List, Any]]:
        if type(obj) == set:
            return sorted(obj)
        return obj

    @staticmethod
    def __filter_vars(variables: List[Tuple[str, Any]]) -> List[Tuple[str, Any]]:
        return [(key, Config.__sort_if_set(value)) for key, value in variables
                if not key.startswith('_')
                and not callable(value)
                and not type(value) == staticmethod]

    def __str__(self) -> str:
        instance_variables: List[Tuple[str, Any]] = sorted([(key, value) for key, value in vars(self).items()])
        instance_variables = self.__filter_vars(instance_variables)
        class_variables: List[Tuple[str, Any]] = sorted([(key, value) for key, value in vars(type(self)).items()])
        class_variables = self.__filter_vars(class_variables)
        lines = [
            pd.DataFrame(data=class_variables, columns=['STATIC VARIABLES', 'VALUES']).to_string(index=False),
            "",
            pd.DataFrame(data=instance_variables, columns=['DYNAMIC VARIABLES', 'VALUES']).to_string(index=False, )
        ]

        return '\n'.join(lines)
